from __future__ import annotations
from typing import TYPE_CHECKING, Type, Optional
from pydantic import Field

from spice_rack import _bases

if TYPE_CHECKING:
    from spice_rack._frozen_registry._base import FrozenRegistryBase


__all__ = (
    "ItemIndexInvalidException",
)


class _ItemIxInvalidErrorInfo(_bases.exceptions.ErrorInfoBase):
    """tried to get an item with an index int, but the int was invalid for this registry"""
    registry_type: str = Field(description="the name of the registry class that made this error")
    invalid_ix: int = Field(description="the index we tried, that was found invalid")
    registry_size: int = Field(description="the total size of this registry")
    caught_exception: _bases.exceptions.WrappedExternalException = Field(
        description="the exception we caught"
    )


class ItemIndexInvalidException(
    _bases.exceptions.CustomExceptionBase[_ItemIxInvalidErrorInfo]
):
    """tried to get an item with an index int, but the int was invalid for this registry"""
    def __init__(
            self,
            invalid_ix: int,
            registry_type: Type[FrozenRegistryBase],
            registry_size: int,
            caught_exception: Exception,
            verbose: bool = True,
            extra_info: Optional[dict] = None
    ):
        detail = (f"__getitem__ failed with index {invalid_ix}. "
                  f"The '{registry_type.__name__}' registry has {registry_size} items.")

        error_info = {
            "registry_type": registry_type.__name__,
            "invalid_ix": invalid_ix,
            "registry_size": registry_size,
            "caught_exception": caught_exception,
        }
        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=verbose,
            extra_info=extra_info
        )



from __future__ import annotations
from uuid import uuid4
import pydantic


__all__ = (
    "GuidT", "GuidTypeAdapter", "gen_guid", "gen_str_guid"
)


GuidT = pydantic.UUID4
"""the pydantic type for UUID4"""

GuidTypeAdapter: pydantic.TypeAdapter[GuidT] = pydantic.TypeAdapter(GuidT)
"""a type adapter that validates strings to 'GuidT' instances"""


def gen_guid() -> GuidT:
    """generate a new UUID4"""
    return uuid4()


def gen_str_guid() -> str:
    """generate a new UUID4 hex"""

    return gen_guid().hex



from __future__ import annotations
import typing as t
import pydantic

from spice_rack._logging._sinks._base import AbstractLogSink

if t.TYPE_CHECKING:
    from loguru import Logger


__all__ = ("FilePathLogSink",)


class FilePathLogSink(AbstractLogSink, class_id="file"):
    """
    see https://loguru.readthedocs.io/en/stable/overview.html
    #easier-file-logging-with-rotation-retention-compression
    for more info. We only support a small subset here for now
    """
    file_name_prefix: str = pydantic.Field(
        description="the prefix of the file name where we'll be writing logs"
    )
    dir_path: pydantic.DirectoryPath
    file_ext: str = "log"
    add_timestamp_to_name: bool = True
    rotation_size: t.Optional[str] = "500 MB"

    def setup(self, logger: Logger, **loguru_kwargs) -> None:
        if not self.dir_path.exists():
            self.dir_path.mkdir(parents=True, exist_ok=True)
        time_str = "{time}"
        file_name = f"{self.name}_{time_str}.{self.file_ext}"
        log_file_path = self.dir_path.joinpath(file_name)
        logger.add(
            sink=str(log_file_path),
            rotation=self.rotation_size,
            **loguru_kwargs
        )



from __future__ import annotations
import types
import typing as t
from typing_extensions import LiteralString
from devtools import pformat
from pydantic_core import PydanticCustomError

from spice_rack._bases._exception import _error_info, _exception_exception

if t.TYPE_CHECKING:
    from spice_rack._bases._exception import _http


__all__ = (
    "CustomExceptionBase",
)


Self = t.TypeVar("Self", bound="CustomExceptionBase")
"""pycharm type checker supports this, but it is same as new 'Self' type ann"""

ErrorInfoTV = t.TypeVar("ErrorInfoTV", bound=_error_info.ErrorInfoBase)


class CustomExceptionBase(Exception, t.Generic[ErrorInfoTV]):
    """
    Base for a custom-built exception class.

    Note:
        the typing for the
    """
    _error_info_cls: t.ClassVar[t.Type[_error_info.ErrorInfoBase]] = None  # type: ignore

    _detail: str
    _verbose: bool
    _error_info_inst: ErrorInfoTV

    def __class_getitem__(cls, item: t.Union[t.TypeVar, t.Type[_error_info.ErrorInfoBase]]):

        error_info_cls: t.Type[_error_info.ErrorInfoBase]

        # if manually set, just keep that
        if cls._error_info_cls is not None:
            error_info_cls = cls._error_info_cls
        else:
            try:
                if isinstance(item, t.TypeVar):
                    # if a type var, make sure we have a bound, and use that
                    if item.__bound__:
                        error_info_cls = item.__bound__
                    else:
                        raise ValueError(
                            f"'{item}' is a type var, and must have a bound set"
                        )

                elif issubclass(item, _error_info.ErrorInfoBase):
                    error_info_cls = item

                else:
                    raise ValueError(f"unable to set error info class from the item, {item}")

            except Exception as e:
                raise ValueError(
                    f"error encountered trying to get error info class from type var, error: {e}"
                ) from e

        intermediate_cls_name = f"{cls.__name__}__{item.__name__}"
        new_cls: t.Type[CustomExceptionBase] = t.cast(
            t.Type[CustomExceptionBase],
            types.new_class(
                intermediate_cls_name, bases=(cls, ),
                kwds={"error_info_cls": error_info_cls}
            )
        )
        return new_cls

    def __init_subclass__(
            cls,
            error_info_cls: t.Optional[t.Type[_error_info.ErrorInfoBase]] = None,
            **kwargs
    ):
        if error_info_cls is None and cls._error_info_cls is None:
            raise ValueError(f"{cls.__name__} must specify error info class'")
        elif error_info_cls is not None:
            cls._error_info_cls = error_info_cls
        super().__init_subclass__(**kwargs)

    def __init__(
            self,
            detail: str,
            error_info: t.Union[ErrorInfoTV, dict],
            verbose: bool = True,
            extra_info: t.Optional[dict] = None
    ):
        self.get_error_info_cls().model_rebuild()

        self._detail = detail
        self._verbose = verbose

        try:
            self._error_info_inst = self.get_error_info_cls().model_validate(error_info)

        except Exception as e:
            raise _exception_exception.CustomExceptionInitializationError(
                detail="error info provided failed validation",
                caught_exception=e
            ) from e

        if extra_info:
            self._error_info_inst.extra.update(extra_info)
        formatted_str: str
        if self.verbose:
            formatted_str = self._get_formatted_str_verbose()
        else:
            formatted_str = self._get_formatted_str_terse()

        super().__init__(formatted_str)

    @property
    def detail(self) -> str:
        return self._detail

    @property
    def verbose(self) -> bool:
        return self._verbose

    @property
    def error_info(self: Self) -> ErrorInfoTV:
        return self._error_info_inst

    @classmethod
    def get_error_info_cls(cls: t.Type[Self]) -> t.Type[ErrorInfoTV]:
        return cls._error_info_cls  # type: ignore

    @classmethod
    @t.final
    def get_error_payload_cls(cls: t.Type[Self]) -> t.Type[_error_info.ErrorPayload[ErrorInfoTV]]:
        """
        Returns: an ErrorPayload class parameterized with the correct
        concrete subclass of ErrorInfoBase
        """
        payload_cls = _error_info.ErrorPayload[cls.get_error_info_cls()]  # type: ignore
        return payload_cls

    @t.final
    def get_error_payload_inst(self: Self) -> _error_info.ErrorPayload[ErrorInfoTV]:
        return _error_info.ErrorPayload[self.get_error_info_cls()](
            detail=self.detail,
            error_type=type(self).__name__,
            info=self.error_info,
        )

    def _get_formatted_str_verbose(self) -> LiteralString:
        return pformat(self.get_error_payload_inst().json_dict(use_str_fallback=True))

    def _get_formatted_str_terse(self) -> LiteralString:
        return t.cast(LiteralString, self.detail)

    def as_http_error_resp(
            self,
            status_code: int
    ) -> _http.HttpErrorResponse[ErrorInfoTV]:
        from spice_rack._bases._exception import _http
        return _http.HttpErrorResponse[self._error_info_cls](
            status_code=status_code,
            error_payload=self.get_error_payload_inst()
        )

    def as_http_error_exc(
            self,
            status_code: int
    ) -> _http.HttpException[ErrorInfoTV]:
        from spice_rack._bases._exception import _http
        return _http.HttpException[self._error_info_cls](
            status_code=status_code,
            error_payload=self.get_error_payload_inst()
        )

    @classmethod
    @t.final
    def get_http_response_info_inst(
            cls,
            status_code: int,
            custom_desc: t.Optional[str] = None
    ) -> _http.HttpErrorResponseInfo:
        """build a HttpErrorResponseInfo inst for this exception class"""
        from spice_rack._bases._exception import _http
        return _http.HttpErrorResponseInfo(
            status_code=status_code,
            exception_type=cls,
            custom_desc=custom_desc
        )

    def as_pydantic_error(self, manual_verbose: t.Optional[bool] = None) -> PydanticCustomError:
        verbose: bool
        if manual_verbose is not None:
            verbose = manual_verbose
        else:
            verbose = self.verbose

        if verbose:
            msg = self._get_formatted_str_verbose()
        else:
            msg = self._get_formatted_str_terse()

        error_type = t.cast(LiteralString, self.__class__.__name__)

        return PydanticCustomError(error_type, msg)



from spice_rack import (
    _bases as bases,
    _logging as logging,
    _fs_ops as fs_ops,
    _gcp_auth as gcp_auth,
    _ts_service as ts_service,
    _misc as misc,
    _utils as utils,
    _guid_service as guid_service,
    _frozen_registry as frozen_registry,
    _polars_service as polars_service,
    _common_validators as common_validators,
    _api_helpers as api_helpers,
    _version_getter
)
GuidT = guid_service.GuidT
Logger = logging.Logger


__version__ = _version_getter.get_version(__name__)
VERSION = __version__



from __future__ import annotations
from abc import abstractmethod
import typing as t
from fsspec.spec import AbstractFileSystem as AbstractFsSpecFileSystem
import pydantic

from spice_rack import _bases, _logging
from spice_rack._fs_ops import _path_strs, _open_modes, _exceptions


__all__ = (
    "AbstractFileSystem",
)


class AbstractFileSystem(
    _bases.dispatchable.DispatchableValueModelBase,
    _logging.log_extra.LoggableObjMixin
):
    """base class for all fsspec file system wrappers"""

    @abstractmethod
    def build_fsspec_file_system(self) -> AbstractFsSpecFileSystem:
        ...

    @property
    def fsspec_obj(self) -> AbstractFsSpecFileSystem:
        return self.build_fsspec_file_system()

    @classmethod
    @abstractmethod
    def get_fs_specific_prefix(cls) -> str:
        ...

    @classmethod
    @pydantic.validate_call
    def clean_raw_path_str(
            cls,
            __raw_path: str,
    ) -> _path_strs.FileOrDirAbsPathT:
        """
        strip the raw path of the file system-specific prefix, and cast it to a
        standardized directory or file path representation.

        This is the inverse of 'contextualize_abs_path'.
        """
        cleaned_path = __raw_path.replace(
            cls.get_fs_specific_prefix(),
            "/",
            1
        )
        return _path_strs.FileOrDirAbsPathTypeAdapter.validate_python(
            cleaned_path
        )

    @pydantic.validate_call
    def contextualize_abs_path(self, __path: _path_strs.FileOrDirAbsPathT) -> str:
        """
        convert a standardized file or directory path representation into a
        str starting with the file system-specific prefix.
        This is the inverse of 'clean_raw_path_str'.
        """
        prefix_val = self.get_fs_specific_prefix()
        formatted_path_str = prefix_val + str(__path)[1:]
        return formatted_path_str

    @abstractmethod
    def get_home_dir(self) -> _path_strs.AbsoluteDirPathStr:
        ...

    def __hash__(self) -> int:
        return hash(
            self.contextualize_abs_path(self.get_home_dir())
        )

    def __eq__(self, other: AbstractFileSystem) -> bool:
        if type(self) == type(other):  # noqa
            if hash(self) == hash(other):
                return True
        return False

    def special_repr(self) -> str:
        return f"FileSystem[{self.class_id}]"

    def __get_logger_data__(self) -> t.Dict:
        return {
            "file_system_type": self.class_id,
            "file_system_home_dir": self.contextualize_abs_path(self.get_home_dir())
        }

    @pydantic.validate_call
    def exists(self, __path: _path_strs.FileOrDirAbsPathT) -> bool:
        """
        returns True if this file system object exists, false otherwise
        """

        # todo: what if perms issue not existence issue?
        return self.fsspec_obj.exists(self.contextualize_abs_path(__path))

    @pydantic.validate_call
    def ensure_exists(self, __path: _path_strs.FileOrDirAbsPathT) -> None:
        """
        ensure the file or dir path exists, raising an exception if not

        Args:
            __path: the path to the file or directory

        Returns: Nothing

        Raises:
            NonExistentPathException: if the path doesn't exist, or we cannot access it

        """
        exists = self.exists(__path)
        if not exists:
            raise _exceptions.NonExistentPathException(
                file_system=self,
                path=__path,
            )
        else:
            return

    @pydantic.validate_call
    def ensure_nonexistent(self, __path: _path_strs.FileOrDirAbsPathT) -> None:
        """
        ensure the file or dir path does exist, raising an exception if it does

        Args:
            __path: the path to the file or directory

        Returns: Nothing

        Raises:
            PathAlreadyExistsException: if the path does exist
        """
        exists = self.exists(__path)
        if exists:
            raise _exceptions.PathAlreadyExistsException(
                file_system=self,
                path=__path,
            )
        else:
            return

    @pydantic.validate_call
    def open_file(
            self,
            __path: _path_strs.AbsoluteFilePathStr,
            mode: _open_modes.SupportedOpenModesT
    ) -> _open_modes.OpenFileT:
        """
        return a readable open file object. todo: revisit type annotations of the return type here

        Args:
            __path: the file path str
            mode: the mode we are opening in

        Returns:
            the readable open file object
        """
        if mode != "wb":
            self.ensure_exists(__path)
        return self.fsspec_obj.open(
            path=self.contextualize_abs_path(__path), mode=mode
        )

    @pydantic.validate_call
    def delete_file(
            self,
            __path: _path_strs.AbsoluteFilePathStr,
            *,
            if_non_existent: t.Literal["raise", "return"] = "return"
    ) -> None:
        """
        delete the file
        """
        exists = self.exists(__path)
        if not exists:
            if if_non_existent == "raise":
                self.ensure_exists(__path)

        else:
            # what if perms issue not existence issue?
            self.fsspec_obj.delete(path=self.contextualize_abs_path(__path), recursive=True)
        return

    @pydantic.validate_call
    def delete_dir(
            self,
            __path: _path_strs.AbsoluteDirPathStr,
            *,
            recursive: bool = True,
            if_non_existent: t.Literal["raise", "return"] = "return"
    ) -> None:
        """
        delete the directory, if recursive is true, we delete all files and subdirectories
        """
        exists = self.exists(__path)
        if not exists:
            if if_non_existent == "raise":
                self.ensure_exists(__path)

        else:
            # what if perms issue not existence issue?
            self.fsspec_obj.delete(path=self.contextualize_abs_path(__path), recursive=recursive)
        return

    @pydantic.validate_call
    def iter_dir_contents(
            self,
            __path: _path_strs.AbsoluteDirPathStr,
    ) -> t.Iterator[_path_strs.FileOrDirAbsPathT]:
        """iterate over the top level dir contents"""
        for raw_info_rec_i in self.fsspec_obj.listdir(
            self.contextualize_abs_path(__path)
        ):
            raw_path_i = raw_info_rec_i.get("name")
            if raw_path_i is None:
                raise ValueError(
                    "'name' not found in record?"
                )
                # raise _exceptions.FileSystemException(
                #     file_system=self,
                #     extra_info={
                #         "dir_path": path,
                #         "raw_record": raw_info_rec_i,
                #         "issue": "'name' not found in record?"
                #     },
                # )

            path_i: _path_strs.FileOrDirAbsPathT
            raw_type = raw_info_rec_i.get("type")
            formatted_path_i = str(self.clean_raw_path_str(raw_path_i))

            if raw_type == "file":
                path_i = _path_strs.AbsoluteFilePathStr(formatted_path_i)

            elif raw_type == "directory":
                if not formatted_path_i.endswith("/"):
                    formatted_path_i = f"{formatted_path_i}/"
                path_i = _path_strs.AbsoluteDirPathStr(formatted_path_i)

            else:
                raise ValueError(f"unexpected type val: '{raw_type}'")

            yield path_i

    @pydantic.validate_call
    def list_dir_contents(
            self,
            __path: _path_strs.AbsoluteDirPathStr
    ) -> list[_path_strs.FileOrDirAbsPathT]:
        """exhaust the iterator built from 'iter_dir_contents' returning
        the items in a list"""
        return list(self.iter_dir_contents(__path))

    @pydantic.validate_call
    def iter_dir_contents_files_only(
            self,
            __path: _path_strs.AbsoluteDirPathStr,
            *,
            recursive: bool = True
    ) -> t.Iterator[_path_strs.AbsoluteFilePathStr]:
        """iterate over every file in a directory, recursing into subdirectories if
        recursive is True"""
        for path_i in self.iter_dir_contents(__path):
            if isinstance(path_i, _path_strs.AbsoluteFilePathStr):
                yield path_i
            elif isinstance(path_i, _path_strs.AbsoluteDirPathStr):
                if recursive:
                    for path_ij in self.iter_dir_contents_files_only(
                        path=path_i,
                        recursive=recursive,
                    ):
                        yield path_ij
                else:
                    continue
            else:
                raise ValueError(type(path_i))

    @pydantic.validate_call
    def make_dir(
            self,
            __path: _path_strs.AbsoluteDirPathStr,
            *,
            if_exists: t.Literal["raise", "return"] = "return",
            create_parents: bool = True
    ) -> None:
        """create a new directory, and parents if the parent directory doesn't exist"""
        if self.exists(__path):
            if if_exists == "raise":
                raise _exceptions.PathAlreadyExistsException(
                    file_system=self,
                    path=__path,
                    extra_info={
                        "action_attempted": "make_dir"
                    }
                )
            else:
                return

        else:
            self.fsspec_obj.mkdir(
                path=self.contextualize_abs_path(__path),
                create_parents=create_parents
            )

    @pydantic.validate_call
    def download_file_locally(
            self,
            __source_path: _path_strs.AbsoluteFilePathStr,
            __local_dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteFilePathStr:
        """
        download file from current file system into the local file system.

        the downloaded file will have the same name as the source file, inside the
        specified local directory.
        """
        from spice_rack._fs_ops._file_systems import _local
        local_fs = _local.LocalFileSystem()
        local_fs.make_dir(__local_dest_dir, if_exists="return", create_parents=True)

        local_path = __local_dest_dir.joinpath(
            _path_strs.RelFilePathStr(__source_path.get_name(include_suffixes=True))
        )
        self.fsspec_obj.download(
            lpath=local_fs.contextualize_abs_path(local_path),
            rpath=self.contextualize_abs_path(__source_path)
        )
        return local_path

    @pydantic.validate_call
    def download_dir_locally(
            self,
            __source_dir: _path_strs.AbsoluteDirPathStr,
            __local_dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteDirPathStr:
        """
        download directory from current file system into the local file system.

        the downloaded directory will have the same name as the source directory, inside the
        specified local directory.
        """
        from spice_rack._fs_ops._file_systems import _local
        from spice_rack._fs_ops._helpers import is_placeholder_file_path
        local_fs = _local.LocalFileSystem()
        local_fs.make_dir(__local_dest_dir, if_exists="return", create_parents=True)

        local_path = __local_dest_dir.joinpath(
            _path_strs.RelDirPathStr(__source_dir.get_name())
        )
        self.fsspec_obj.download(
            lpath=local_fs.contextualize_abs_path(local_path),
            rpath=self.contextualize_abs_path(__source_dir),
            recursive=True
        )

        # remove possible placeholder files
        for local_path_i in local_fs.iter_dir_contents_files_only(local_path):
            if is_placeholder_file_path(local_path_i):
                local_fs.delete_file(
                    local_path_i, if_non_existent="return"
                )
        return local_path



from __future__ import annotations
import typing as t
import pydantic
from google.auth.credentials import Credentials  # noqa
from google.oauth2 import service_account  # noqa

from spice_rack._gcp_auth._auth_strategies import _base
from spice_rack import _bases


__all__ = (
    "ServiceAcctKeyFileAuthStrategy",
    "GcpServiceAccountKeyData",
    "ServiceAcctKeyAuthStrategy"
)


class ServiceAcctKeyFileAuthStrategy(_base.AbstractGcpAuthStrategy):
    """
    see `docs <https://googleapis.dev/python/google-api-core/latest/auth.html#authentication>`_
    """
    file_path: str = pydantic.Field(
        description="the raw string to a given file that holds the creds"
    )

    def build_gcp_creds(self) -> Credentials:
        return service_account.Credentials.from_service_account_file(
            filename=self.file_path
        )


class GcpServiceAccountKeyData(_bases.value_model.ValueModelBase):
    """data from a gcp service account's json key"""
    type: str
    project_id: str
    private_key_id: str
    private_key: str
    client_email: str
    client_id: str
    auth_uri: str
    token_uri: str
    auth_provider_x509_cert_url: str
    client_x509_cert_url: str
    universe_domain: t.Optional[str] = None
    refresh_token: t.Optional[str] = None
    client_secret: t.Optional[str] = None

    @pydantic.field_validator("private_key", mode="before")
    def _remove_double_new_lines(cls, v: str):
        # this happens when we json encode this object sometimes IDK why
        v = v.replace("\\n", "\n")
        return v


class ServiceAcctKeyAuthStrategy(_base.AbstractGcpAuthStrategy):
    """
    see `docs <https://googleapis.dev/python/google-api-core/latest/auth.html#authentication>`_
    """
    key_data: GcpServiceAccountKeyData = pydantic.Field(
        description="the actual json contents of the key",
    )

    def build_gcp_creds(self) -> Credentials:
        return service_account.Credentials.from_service_account_info(
            info=self.key_data.model_dump(exclude_defaults=True)
        )



from __future__ import annotations
import typing as t
import pydantic

from spice_rack import _bases

__all__ = (
    "LogLevel",
)


_LogLevelStrLiteralT = t.Literal[
    "debug",
    "info",
    "warning",
    "error",
    "critical"
]
_LogLevelIntLiteralT = t.Literal[
    10,
    20,
    30,
    40,
    50
]


_LogLevelLiteralT = t.Union[_LogLevelStrLiteralT, _LogLevelIntLiteralT]

_str_to_int: t.Dict[_LogLevelStrLiteralT, _LogLevelIntLiteralT] = dict(
    zip(t.get_args(_LogLevelStrLiteralT), t.get_args(_LogLevelIntLiteralT))
)
_int_to_str: t.Dict[_LogLevelIntLiteralT, _LogLevelStrLiteralT] = dict(
    zip(t.get_args(_LogLevelIntLiteralT), t.get_args(_LogLevelStrLiteralT))
)


class LogLevel(_bases.root.RootModel[_LogLevelLiteralT]):
    """
    a root model holding enum values.
    """
    _str_to_int: t.ClassVar[t.Dict[_LogLevelStrLiteralT, _LogLevelIntLiteralT]] = _str_to_int
    _int_to_str: t.ClassVar[t.Dict[_LogLevelIntLiteralT, _LogLevelStrLiteralT]] = _int_to_str

    @pydantic.model_validator(mode="before")
    def _ensure_lowercase(cls, data: t.Any) -> t.Any:
        if isinstance(data, str):
            if not data.islower():
                data = data.lower()
        return data

    def __str__(self) -> _LogLevelStrLiteralT:
        str_val: _LogLevelStrLiteralT
        if isinstance(self.root, str):
            # must be this type
            str_val = t.cast(_LogLevelStrLiteralT, self.root)
        else:
            root_data: _LogLevelIntLiteralT = self.root
            if root_data not in self._int_to_str:
                raise ValueError(
                    f"{root_data} doesn't have a corresponding mapping to a str log level"
                )
            else:
                str_val = self._int_to_str[root_data]

        return str_val

    def __int__(self) -> _LogLevelIntLiteralT:
        int_val: _LogLevelIntLiteralT
        if isinstance(self.root, int):
            # must be one of the int literal options
            int_val = t.cast(_LogLevelIntLiteralT, self.root)

        else:
            root_data: _LogLevelStrLiteralT = self.root
            if root_data not in self._str_to_int:
                raise ValueError(
                    f"'{root_data}' doesn't have a corresponding mapping to a int log level"
                )
            else:
                int_val = self._str_to_int[root_data]

        return int_val



from __future__ import annotations
from typing import TYPE_CHECKING, Type, Optional
from pydantic import Field

from spice_rack import _bases

if TYPE_CHECKING:
    from spice_rack._frozen_registry._base import FrozenRegistryBase

__all__ = (
    "KeyNotFoundException",
)


class _KeyNotFoundErrorInfo(_bases.exceptions.ErrorInfoBase):
    """the specified key wasn't found in the registry"""
    registry_type: str = Field(description="the name of the registry class that made this error")
    missing_key: str = Field(description="the key we tried to find")
    keys_in_registry: list[str] = Field(description="the keys that are present")


class KeyNotFoundException(
    _bases.exceptions.CustomExceptionBase[_KeyNotFoundErrorInfo]
):
    """the specified key wasn't found in the registry"""
    def __init__(
            self,
            missing_key: str,
            registry_type: Type[FrozenRegistryBase],
            keys_in_registry: list[str],
            verbose: bool = True,
            extra_info: Optional[dict] = None
    ):
        detail = (f"failed to find the key, '{missing_key}', "
                  f"in the '{registry_type.__name__}' registry")

        error_info = {
            "registry_type": registry_type.__name__,
            "missing_key": missing_key,
            "keys_in_registry": keys_in_registry,
        }
        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=verbose,
            extra_info=extra_info
        )



from __future__ import annotations
import typing as t
import typing_inspect
import pydantic
from pydantic import Field
import functools

from spice_rack import _bases, _utils
from spice_rack._frozen_registry import _exceptions

__all__ = (
    "FrozenRegistryMember",
    "FrozenRegistryBase",
)


# Note: the "_RegistryItemTV" class must have the specified key_attr_name attribute,
# but we can't specify this via base class, so we'll just manually check for it here.
_RegistryItemTV = t.TypeVar("_RegistryItemTV", bound=_bases.ValueModelBase)
_RegistryItemKeyTV = t.TypeVar("_RegistryItemKeyTV", bound=str)
Self = t.TypeVar("Self", bound="FrozenRegistryBase")


class FrozenRegistryMember(_bases.ValueModelBase, t.Generic[_RegistryItemKeyTV, _RegistryItemTV]):
    """a member of a frozen registry, holding a key and an item"""
    key: _RegistryItemKeyTV = Field(description="the key we use to look up the item")
    item: _RegistryItemTV = Field(
        description="The item itself. The frozen registry is meant to simulate we have a collection of these items,"
                    " with some extra built-in functionality and safeties"
    )


class FrozenRegistryBase(
    _bases.RootModel[
        t.Tuple[FrozenRegistryMember[_RegistryItemKeyTV, _RegistryItemTV], ...]
    ],
    t.Generic[_RegistryItemTV, _RegistryItemKeyTV]
):
    """
    Base class for a registry of items, where we can control how we get and add things to it.
    We use this pattern in various places, so consolidating behavior to a base class here
    minimizes the small code drift between the different implementations.
    """
    _key_lookup: t.Dict[_RegistryItemKeyTV, int] = pydantic.PrivateAttr(default=None)

    @classmethod
    @t.final
    def _get_member_cls(cls) -> t.Type[FrozenRegistryMember[_RegistryItemKeyTV, _RegistryItemTV]]:
        root_ann = cls.model_fields["root"].annotation
        member_cls: t.Type[FrozenRegistryMember[_RegistryItemKeyTV, _RegistryItemTV]] = t.get_args(root_ann)[0]
        return member_cls

    @classmethod
    @t.final
    def is_item_cls_set(cls) -> bool:
        member_cls = cls._get_member_cls()
        item_ann = member_cls.model_fields["item"].annotation

        return not typing_inspect.is_typevar(item_ann)

    @classmethod
    @t.final
    def is_key_cls_set(cls) -> bool:
        member_cls = cls._get_member_cls()
        key_ann = member_cls.model_fields["key"].annotation
        return not typing_inspect.is_typevar(key_ann)

    @classmethod
    @t.final
    def get_item_cls(cls) -> t.Type[_RegistryItemTV]:
        """Returns the type specified, if it is a type var, this raises a ValueError"""
        member_cls = cls._get_member_cls()
        item_ann = member_cls.model_fields["item"].annotation

        if typing_inspect.is_typevar(item_ann):
            raise ValueError(
                f"'{item_ann}' is still a type var and not parameterized yet"
            )
        else:
            return item_ann

    @classmethod
    @t.final
    def get_key_cls(cls) -> t.Type[_RegistryItemKeyTV]:
        member_cls = cls._get_member_cls()
        key_ann = member_cls.model_fields["key"].annotation

        if typing_inspect.is_typevar(key_ann):
            raise ValueError(
                f"'{key_ann}' is still a type var and not parameterized yet"
            )
        else:
            return key_ann

    @classmethod
    def _key_getter_fn_spec(cls) -> t.Union[t.Callable[[_RegistryItemTV], _RegistryItemKeyTV], str]:
        """
        Convenience method to allow for specifying an attr as the getter easily, while keeping our error messaging
        in the '_key_getter_func_builder'.

        overwrite this to customize how get the key from the item.
        This is used to automatically parse '_RegistryItemTV' instances into 'FrozenRegistryMember' instances

        If this returns a string, we assume it an attr access, i.e. if it returns "key" the function used will
        be getattr(item, "key"). If callable, we'll just use the callable directly
        """
        return "key"

    @classmethod
    @functools.lru_cache(maxsize=1)
    @t.final
    def _key_getter_fn_builder(cls) -> t.Callable[[_RegistryItemTV], _RegistryItemKeyTV]:
        """this returns the actual function we'll use to get the key"""
        getter_fn = cls._key_getter_fn_spec()

        final_fn: t.Callable[[_RegistryItemTV], _RegistryItemKeyTV]
        if isinstance(getter_fn, str):
            attr_name = getter_fn
            # make sure either a property or a pydantic field
            item_cls: pydantic.BaseModel = cls.get_item_cls()

            if attr_name not in item_cls.model_fields and not hasattr(item_cls, attr_name):
                raise ValueError(
                    f"attr name specified, '{attr_name}', not found on the item cls {item_cls} "
                )

            def final_fn(__item: _RegistryItemTV) -> _RegistryItemKeyTV:
                return getattr(__item, getter_fn)
        else:
            final_fn = getter_fn

        return final_fn

    @classmethod
    def validate_class_setup(cls) -> None:
        # make sure we can build this key getter function
        cls._key_getter_fn_builder()

    def __init_subclass__(cls, **kwargs):
        # if base we don't validate
        if cls.is_key_cls_set() and cls.is_item_cls_set():
            cls.validate_class_setup()

        super().__init_subclass__(**kwargs)

    @classmethod
    def _item_to_member(cls, __item: t.Any) -> FrozenRegistryMember[_RegistryItemKeyTV, _RegistryItemTV]:
        item = cls.get_item_cls().model_validate(__item)
        key = cls._key_getter_fn_builder()(item)
        return FrozenRegistryMember[cls.get_key_cls(), cls.get_item_cls()](
            key=key,
            item=item
        )

    # noinspection PyNestedDecorators
    @pydantic.field_validator("root", mode="before")
    @classmethod
    def _auto_parse_items(cls, data: t.Any) -> t.Any:
        # parse a raw list or tuple
        if isinstance(data, (list, tuple)):
            members: t.List[FrozenRegistryMember] = []
            for obj in data:

                # if already a registry member class, we don't parse just use it
                if isinstance(obj, FrozenRegistryMember):
                    members.append(obj)

                # if a dict, we check if it has the same 2 fields as FrozenRegistryMember
                # and if so we try to parse it that way first. If it fails we see if it is an
                # item with those 2 fields too.
                # If it has other fields, we parse it as an item
                elif isinstance(obj, dict):
                    if set(obj.keys()) == {"key", "item"}:
                        # try to parse as member first
                        try:
                            member = cls._get_member_cls().model_validate(obj)
                        except pydantic.ValidationError:
                            member = cls._item_to_member(obj)
                        except Exception as e:
                            raise e

                    else:
                        member = cls._item_to_member(obj)

                    members.append(member)

                # we parse it as an item directly
                else:
                    member = cls._item_to_member(obj)
                    members.append(member)

            data = members
        return data

    @pydantic.model_serializer(mode="plain", when_used="json")
    def _json_serialize(self) -> t.List[t.Dict[str, pydantic.JsonValue]]:
        return [item.model_dump(mode="json") for item in self.iter_items()]

    @classmethod
    def _get_default_value(cls) -> tuple[FrozenRegistryMember, ...]:
        return ()

    @classmethod
    def init_empty(cls: t.Type[Self]) -> Self:
        return cls(root=())

    def _post_init_setup(self) -> None:
        self._key_lookup = {
            member.key: member_ix for member_ix, member in enumerate(self.iter_members())
        }

    def size(self) -> int:
        return len(self.root)

    def iter_members(self) -> t.Iterator[FrozenRegistryMember[_RegistryItemTV, _RegistryItemKeyTV]]:
        for member in self.root:
            yield member

    def iter_items(self: Self) -> t.Iterator[_RegistryItemTV]:
        """iterator over all items"""
        for member in self.iter_members():
            yield member.item

    def __len__(self) -> int:
        return self.size()

    def __getitem__(self, ix: int) -> _RegistryItemTV:
        try:
            return self.root[ix].item
        except Exception as e:
            raise _exceptions.ItemIndexInvalidException(
                invalid_ix=ix,
                registry_type=type(self),
                registry_size=self.size(),
                caught_exception=e,
                verbose=True,
                extra_info={}
            )

    def __iter__(self: Self) -> t.Iterator[_RegistryItemTV]:
        for item in self.iter_items():
            yield item

    def list_keys(self: Self) -> list[_RegistryItemKeyTV]:
        """
        Returns: list all keys in the registry
        """
        return list(self._key_lookup.keys())

    def get_item_maybe(self, __key: _RegistryItemKeyTV) -> t.Optional[_RegistryItemTV]:
        """
        get an item for the key or return None if nothing is found

        Args:
            __key: the key we're looking for

        Returns: an instance of _RegistryItemTV or None
        """
        __key = self.get_key_cls()(__key)
        if __key in self._key_lookup:
            member_ix = self._key_lookup[__key]
            return self.root[member_ix].item
        else:
            return

    def get_item(self, __key: _RegistryItemKeyTV) -> _RegistryItemTV:
        """
        get an item for the key, or raise an exception

        Args:
            __key: the key we're looking for

        Returns: an instance of _RegistryItemTV

        Raises:
            KeyNotFoundException: if nothing is found
        """
        item_maybe = self.get_item_maybe(__key)
        if item_maybe is None:
            raise _exceptions.KeyNotFoundException(
                missing_key=__key,
                registry_type=type(self),
                keys_in_registry=self.list_keys(),
                extra_info={}
            )
        else:
            return item_maybe

    def key_exists(self, __key: _RegistryItemKeyTV) -> bool:
        return __key in self._key_lookup

    def raise_if_found(self, __key: _RegistryItemKeyTV) -> None:
        """
        convenience method that raises an exception if the key is found

        Raises:
            KeyAlreadyExistsException: if key is found
        """
        if __key in self._key_lookup:
            raise _exceptions.KeyAlreadyExistsException(
                duplicate_key=__key,
                registry_type=type(self),
                extra_info={}
            )

    def raise_if_not_found(self, __key: _RegistryItemKeyTV) -> None:
        """
        convenience method that raises an exception if the key is not found

        Raises:
            KeyNotFoundException: if key is found
        """
        if __key not in self._key_lookup:
            raise _exceptions.KeyNotFoundException(
                missing_key=__key,
                registry_type=type(self),
                keys_in_registry=self.list_keys(),
                extra_info={}
            )

    def with_new_items(
            self: Self,
            new_items: list[_RegistryItemTV],
            if_exists: t.Literal["raise", "replace", "skip"] = "replace"
    ) -> Self:
        """
        create a new instance of this registry with the new items, raising an exception if
        an item already exists with the same key and 'if_exists' equals 'raise'

        Args:
            new_items: the list of items we are adding
            if_exists: if 'raise' we raise an exception if an item already exists with this key,
                and if 'replace', we replace the existing item with the new item,
                if "skip", we don't add the new item, but don't raise an error either

        Returns: a new instance of this class

        Raises:
            KeyAlreadyExistsException: an item with the same key as the new item already exists
                and 'if_exists' is set to 'raise'
        """

        # remake the key set as we add each item, so duplicates within the group of new
        # items don't make their way in.
        curr_members: dict[_RegistryItemKeyTV, FrozenRegistryMember[_RegistryItemKeyTV, _RegistryItemTV]] = {
            member.key: member for member in self.iter_members()
        }
        new_members: t.List[FrozenRegistryMember] = [
            self._item_to_member(item) for item in new_items
        ]

        for new_member_i in new_members:
            if new_member_i.key in curr_members:
                if if_exists == "raise":
                    raise _exceptions.KeyAlreadyExistsException(
                        duplicate_key=new_member_i.key,
                        registry_type=type(self),
                        extra_info={}
                    )
                elif if_exists == "replace":
                    curr_members[new_member_i.key] = new_member_i

                elif if_exists == "skip":
                    continue

                else:
                    raise ValueError(f"unexpected value for 'if_exists': '{if_exists}'")
            else:
                curr_members[new_member_i.key] = new_member_i

        return t.cast(
            Self, self.__class__(tuple(curr_members.values()))
        )

    def without_items(
            self: Self,
            keys: list[_RegistryItemKeyTV],
            if_not_found: t.Literal["raise", "continue"] = "continue"
    ) -> Self:
        """
        create a new registry, without the item tied to the key specified

        Args:
            keys: the list of keys to the items we are deleting
            if_not_found: if 'raise', we raise an error if nothing exists with a key,
                and if 'continue' we just move forward as if we deleted it.

        Returns: the new instance of Self

        Raises:
            KeyNotFoundException: if nothing is found for a key, and 'if_not_found' is 'raise'
        """

        # make a set of keys to check against
        keys_to_delete: set[_RegistryItemKeyTV] = set()
        for key_i in keys:
            if key_i not in self._key_lookup:
                if if_not_found == "raise":
                    # will raise for us
                    self.raise_if_not_found(key_i)
                else:
                    continue
            else:
                keys_to_delete.add(key_i)

        updated_members: list[FrozenRegistryMember] = []
        for member in self.iter_members():
            if member.key in keys_to_delete:
                continue
            else:
                updated_members.append(member)
        res = type(self)(tuple(updated_members))
        return t.cast(Self, res)

    def with_new_item(
            self: Self,
            new_item: _RegistryItemTV,
            if_exists: t.Literal["raise", "replace", "skip"] = "replace"
    ) -> Self:
        """
        create a new instance of this registry with a new item, raising an exception if
        an item already exists with the same key and 'if_exists' equals 'raise'

        Args:
            new_item: the item we are adding
            if_exists: if 'raise' we raise an exception if an item already exists with this key,
                and if 'replace', we replace the existing item with the new item,
                if "skip", we don't add the new item, but don't raise an error either

        Returns: a new instance of this class

        Raises:
            KeyAlreadyExistsException: an item with the same key as the new item already exists
                and 'if_exists' is set to 'raise'
        """
        return self.with_new_items(new_items=[new_item], if_exists=if_exists)

    def without_item(
            self: Self,
            key: _RegistryItemKeyTV,
            if_not_found: t.Literal["raise", "continue"] = "continue"
    ) -> Self:
        """
        create a new registry, without the item tied to the key specified

        Args:
            key: key to the item we are deleting
            if_not_found: if 'raise', we raise an error if nothing exists with this key,
                and if 'continue' we just move forward as if we deleted it.

        Returns: the new instance of Self

        Raises:
            KeyNotFoundException: if nothing is found for the key, and 'if_not_found' is 'raise'
        """
        return self.without_items(keys=[key], if_not_found=if_not_found)

    def __eq__(self: Self, other: t.Any) -> bool:
        if isinstance(other, type(self)):

            # if different size, we know not equal
            if self.size() != other.size():
                return False

            # check same keys
            self_keys = set(self.list_keys())
            other_keys = set(other.list_keys())

            # extra and missing from the POV of self,
            #   i.e. extra_keys = keys in "self" not in "other"
            extra_keys, missing_keys = _utils.get_difference_between_sets(
                self_keys, other_keys
            )
            if extra_keys or missing_keys:
                return False

            # check each object individually
            for key in self_keys:
                self_obj = self.get_item(key)
                other_obj = other.get_item(key)
                if not self_obj == other_obj:
                    return False

            # if we make it here, they are equal
            return True
        else:

            # not same type, so definitely False
            return False

    def special_repr(self: Self) -> str:
        return (f"{self.__class__.__name__}"
                f"[key_field='{self.get_key_attr_name()}', size={self.size()}]")



from __future__ import annotations
import typing as t
from copy import copy

import pydantic
import pydantic_core

__all__ = (
    "SpecialStrBase",
    "StrKeyBase"
)


# coercible type annotation specifying you could also initialize this class with.
_CoercibleTV = t.TypeVar("_CoercibleTV", )
SelfTV = t.TypeVar("SelfTV", bound="SpecialStrBase")


class SpecialStrBase(str):
    """
    makes it convenient to subclass str to let us signify a given str is a special
    type, like an ID or key. This helps type checking and also allows us to add formatting and
    validation. This class also hooks into pydantic validation automatically.
    """
    @classmethod
    def _parse_non_str(cls, root_data: t.Any) -> str:
        """
        Overwrite this to allow the special str class to parse non-string input.
        Manually check the root_data type to determine if you want to parse, it and return
        super()._parse_non_str if you want to avoid handling the data type.
        """
        raise TypeError(f"'{cls.__name__}' cannot parse data of type {type(root_data)}")

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        """overwrite this to specify the custom str formatting for this class"""
        return root_data

    @classmethod
    def _validate_str_val(cls, root_data: str) -> str:
        """validate that the formatted string data is valid"""
        return root_data

    @classmethod
    def _validate(cls, val: t.Any) -> str:
        str_val: str
        if not isinstance(val, str):
            str_val = cls._parse_non_str(val)
        else:
            str_val = val
        str_val = cls._format_str_val(str_val)
        return cls._validate_str_val(str_val)

    def __new__(cls, v: t.Union[str, _CoercibleTV]):
        """create a new instance of this special type class, validating the data first"""
        v = cls._validate(v)
        return super().__new__(cls, v)

    @classmethod
    def __get_pydantic_core_schema__(
            cls, _: t.Any, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        final_schema = pydantic_core.core_schema.no_info_after_validator_function(
            function=cls, schema=handler(str)
        )
        return final_schema

    def special_repr(self) -> str:
        """
        Gives us a special representation indicating the instance is a special string, not just
        a normal one.

        This keeps the regular '__repr__' method is the same as a normal str instance
        """
        return f"{self.__class__.__name__}['{str(self)}']"

    def __eq__(self: SelfTV, __value: object) -> bool:
        try:
            other_formatted = self._validate(__value)
            return str(self) == str(other_formatted)

        except Exception:
            return False

    def __hash__(self) -> int:
        return hash(str(self))


class StrKeyBase(SpecialStrBase):
    """
    extension on 'AbstractSpecialStr' to easily make a new special str class we intend
    to use as some sort of key. The main functionality is when you want to standardize a str value
    as either upper or lower case, and replace common separators with a standard one.
    """
    @classmethod
    def _get_separator(cls) -> str:
        """the standard separator we want to use for the key. overwrite this to customize it"""
        return "_"

    @classmethod
    def _get_separators_to_replace(cls) -> list[str]:
        """
        a list of common separators we want to replace with the standard separator.
        overwrite this to customize the list.
        """
        return [" ", "_", ".", "-"]

    @classmethod
    def _get_case_formatting(cls) -> _CaseFormattingT:
        """
        determine how to handle the case of the string.

        Returns:
            "upper": cast str to all uppercase
            "lower" cast str to all lowercase
            "as_is": leave string unchanged
        """
        return "as_is"

    @classmethod
    def _apply_case_formatting(cls, s: str) -> str:
        option = cls._get_case_formatting()
        if option == "as_is":
            return s
        elif option == "lower":
            return s.lower()
        elif option == "upper":
            return s.upper()
        else:
            raise ValueError(
                f"'{option}' is not a valid choice for the case formatter"
            )

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        # replace separators
        formatted_str = copy(root_data)

        good_sep = cls._get_separator()
        seps_to_replace = [sep for sep in cls._get_separators_to_replace() if sep != good_sep]
        for bad_sep in seps_to_replace:
            formatted_str = formatted_str.replace(bad_sep, good_sep)

        return cls._apply_case_formatting(formatted_str)


_CaseFormattingT = t.Literal["upper", "lower", "as_is"]



from __future__ import annotations

try:
    from google.auth import credentials  # noqa
except ImportError:
    raise ValueError(
        "cannot use the gcp auth module bc google.auth isn't installed"
    )

from spice_rack import _bases
from spice_rack._gcp_auth import _auth_strategies


__all__ = (
    "AnyGcpAuthStrat",
)


class AnyGcpAuthStrat(
    _bases.dispatchable.DispatchedClassContainer[_auth_strategies.AbstractGcpAuthStrategy]
):
    """container class that holds the different types of gcp auth strategies"""
    def build_gcp_creds(self) -> credentials.Credentials:
        """
        builds the GCP credentials object, based on this strategy.

        Returns: The GCP Credentials instance
        """
        return self.root.build_gcp_creds()

    @classmethod
    def init_default(cls) -> AnyGcpAuthStrat:
        strat = _auth_strategies.DefaultAuthStrategy()
        return AnyGcpAuthStrat.model_validate(strat)



import typing as t
import pydantic

from spice_rack._fs_ops._path_strs._base import *
from spice_rack._fs_ops._path_strs._abs import *
from spice_rack._fs_ops._path_strs._rel import *


_TagsT = t.Literal["rel", "abs"]


def _discrim(raw_str: str) -> _TagsT:
    if raw_str.startswith("/"):
        return "abs"
    else:
        return "rel"


RelOrAbsFilePathT = t.Annotated[
    t.Union[
        t.Annotated[AbsoluteFilePathStr, pydantic.Tag("abs")],
        t.Annotated[RelFilePathStr, pydantic.Tag("rel")],
    ],
    pydantic.Discriminator(_discrim)
]
RelOrAbsFilePathTypeAdapter: pydantic.TypeAdapter[RelOrAbsFilePathT] = pydantic.TypeAdapter(
    RelOrAbsFilePathT
)


RelOrAbsDirPathT = t.Annotated[
    t.Union[
        t.Annotated[AbsoluteDirPathStr, pydantic.Tag("abs")],
        t.Annotated[RelDirPathStr, pydantic.Tag("rel")],
    ],
    pydantic.Discriminator(_discrim)
]


RelOrAbsDirPathTypeAdapter: pydantic.TypeAdapter[RelOrAbsDirPathT] = pydantic.TypeAdapter(
    RelOrAbsDirPathT
)



from __future__ import annotations
import mimetypes
import typing as t

from spice_rack import _bases


__all__ = (
    "FileExt",
    "MimeType"
)


class FileExt(_bases.special_str.SpecialStrBase):
    """formats a file ext to make working with them more convenient"""

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        if root_data.startswith("."):
            root_data = root_data[1:]
        return root_data

    def with_dot_prefix(self) -> str:
        return f".{self}"

    def __eq__(self, other: str) -> bool:
        return str(self) == str(FileExt(other))

    def get_mime_type(self) -> t.Optional[MimeType]:
        try:
            return MimeType.from_file_ext(self)
        except ValueError:
            return None

        except Exception as e:
            raise e


_FILE_EXTENSIONS_SET = set(mimetypes.types_map.keys())
_MIME_TYPES_SET = set(mimetypes.types_map.values())


class MimeType(_bases.special_str.SpecialStrBase):
    """
    a subclass of a string representing a MIME type.
    see: https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types
    """

    @classmethod
    def _format_str(cls, root_data: str) -> str:
        if root_data not in _MIME_TYPES_SET:
            raise ValueError(
                f"'{root_data}' not a valid mime type. see: "
                f"https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types"
            )

        else:
            split = root_data.split("/")
            if len(split) != 2:
                raise ValueError(
                    f"expected mimetype to be in format <type>/<subtype>, '{root_data}' is not."
                )
            return root_data

    @property
    def type(self) -> str:
        """get the top level mimetype"""
        return self.split("/")[0]

    @property
    def subtype(self) -> str:
        """get the subtype for this mimetype"""
        return self.split("/")[-1]

    @classmethod
    def from_file_ext(cls, file_ext_or_dot_file_ext: t.Union[str, FileExt]) -> MimeType:
        """
        get the relevant mimetype for the given file extension

        Args:
            file_ext_or_dot_file_ext: a file extension, either with or without the initial '.'

        Returns: the mime type if it is found, value error if nothing found
        """
        if file_ext_or_dot_file_ext.startswith("."):
            file_ext = file_ext_or_dot_file_ext[1:]
        else:
            file_ext = file_ext_or_dot_file_ext

        # make placeholder path-like str for the stdlib lookup func.
        # https://docs.python.org/3/library/mimetypes.html#mimetypes.guess_type
        path_like = f"x.{file_ext}"
        mime_type, content_encoding = mimetypes.guess_type(url=path_like, strict=True)
        if mime_type is None:
            raise ValueError(
                f"failed to find mimetype for file ext: '{file_ext_or_dot_file_ext}'"
            )
        else:
            return MimeType(mime_type)

    def get_possible_file_extensions(self) -> list[FileExt]:
        """return the list of file extensions for this file mimetype"""
        res = []
        for file_ext, mime_type in mimetypes.types_map.items():
            if mime_type == str(self):
                res.append(FileExt(file_ext))
        return res



from __future__ import annotations
import typing as t

import pydantic


__all__ = (
    "BASE_MODEL_CONFIG",
    "PydanticBase"
)

BASE_MODEL_CONFIG = pydantic.ConfigDict(
    extra="forbid",
)


SelfTV = t.TypeVar("SelfTV", bound="PydanticBase")


class PydanticBase(pydantic.BaseModel):
    """a base model we use for all pydantic models, even the other bases"""

    @classmethod
    def get_cls_name(cls) -> str:
        return cls.__name__

    def _post_init_setup(self) -> None:
        """
        Overwrite this hook to set perform misc. logic before calling _post_init_validation.
        Common use case for this is initializing private attributes.
        Notes: Make sure to call super()._post_init_setup
        """
        return

    def _post_init_validation(self) -> None:
        """
        Overwrite this hook to set perform misc. validation logic on the instance, after
        all other common_validators have been executed.
        This is also called after the '_post_init_setup' hook is called
        Notes: Make sure to call super()._post_init_validation
        """
        return

    @pydantic.model_validator(mode="after")
    def _pydantic_post_init_val_hook(self: SelfTV) -> SelfTV:
        """
        Pydantic's hook that gets executed after we initialize an instance.
        rather than overwrite this, overwrite the '_post_init_setup' and '_post_init_validation'
        hooks
        """
        from spice_rack._bases import _exception

        self._post_init_setup()
        try:
            self._post_init_validation()
        except _exception.CustomExceptionBase as e:
            raise e.as_pydantic_error() from e
        except Exception as e:
            raise e

        return self

    def json_dict(
            self,
            use_str_fallback: bool = True,
            **pydantic_kwargs
    ) -> dict:
        """
        Converts a pydantic instance to a dict, going through json first.
        This is a convenience function for when you want a dict, but want to use the json
        encoders connected to the pydantic object.

        Args:
            use_str_fallback: if true, we cast anything we cannot encode to a string
            **pydantic_kwargs: any kwargs available for pydantic's json method. see their docs

        Returns:
            dict: a natively json-encodeable dict

        Examples:
            simple class with custom date encoder::

                class Data(PydanticBase):
                    class Config:
                        json_encoders = {
                            dt.date: lambda date_obj: date_obj.strftime("%d/%m/%Y")
                        }
                    x: int
                    date: dt.date

                inst = Data(x=1, date=dt.date(year=2023, month=1, day=1))

                # regular dict
                inst.dict() == {"x": 1, "date": dt.date(year=2023, month=1, day=1)}

                # this method
                inst.json_dict() == {"x": 1, "date": "01/01/2023"}
                # note that the date is the json encoding tied to the encoder for this pydantic
                # class. I just used a familiar string format, it could be whatever is specified
        """
        pydantic_kwargs = pydantic_kwargs if pydantic_kwargs is not None else {}
        if use_str_fallback:
            fallback = str
        else:
            fallback = None
        return self.__pydantic_serializer__.to_python(
            self,
            mode="json",
            fallback=fallback,
            **pydantic_kwargs,
        )

    @classmethod
    def _import_forward_refs(cls) -> dict:
        """
        hook provided for convenience when you define a subclass with forward refs,
        to avoid circular import issue. Overwrite this method to do the imports and
        add the newly imported objects to a dict.

        Notes:
            - If you're not sure what to import and what keys to use, look at the stuff
              in the `if TYPE_CHECKING:` block, paste those exact imports in here and
              set them equal to their variable name within your current module's context.

        Returns:
            dict: a dict of references to update

        """
        return {}

    @classmethod
    def model_rebuild(
            cls,
            *,
            force: bool = False,
            raise_errors: bool = True,
            _parent_namespace_depth: int = 2,
            _types_namespace: dict[str, t.Any] | None = None,
    ) -> t.Optional[bool]:
        """
        This extends pydantic's builtin hook for updating forward refs.
        We call our special '_import_forward_refs' hook to automatically bring in the things
        imported there, but also if you want to use it like pydantic, i.e. call it directly and
        provide your refs to update via kwargs, that is still supported.

        Returns:
            None: the schema is already "complete" and rebuilding was not required.
            True: rebuilding was required and was successful
            False: rebuilding was required and not successful
        """
        _types_namespace_dict: dict[str, t.Any] = _types_namespace if _types_namespace else {}
        imported_refs = cls._import_forward_refs()
        _types_namespace_dict.update(imported_refs)
        return super().model_rebuild(
            force=force,
            raise_errors=raise_errors,
            _types_namespace=_types_namespace_dict,
            _parent_namespace_depth=_parent_namespace_depth
        )  # type: ignore



from __future__ import annotations

from devtools import pformat

__all__ = (
    "CustomExceptionInitializationError",
)


class CustomExceptionInitializationError(Exception):
    """
    this is raised when we encounter an exception in our custom exception class

    Notes:
        - This is NOT a subclass of our custom exception base class, so if we have a problem
          with our base class we won't end up in recursive spiral
    """

    def __init__(
            self,
            detail: str,
            caught_exception: Exception
    ):

        data_dict = {
            "detail": detail,
            "exception_type": type(caught_exception).__name__,
            "exception_detail": str(caught_exception)
        }
        super().__init__(pformat(data_dict))



from __future__ import annotations
from abc import abstractmethod

from google.auth import credentials  # noqa

from spice_rack import _bases


__all__ = (
    "AbstractGcpAuthStrategy",
)


class AbstractGcpAuthStrategy(
    _bases.dispatchable.DispatchableValueModelBase,
    class_type="root",
    class_id="gcp_auth_strat_base"
):
    """
    base class for all the ways we can authenticate to gcs.
    see `docs <https://googleapis.dev/python/google-api-core/latest/auth.html#authentication>`_
    """

    @abstractmethod
    def build_gcp_creds(self) -> credentials.Credentials:
        """
        builds the GCP credentials object, based on this strategy.

        Returns: The GCP Credentials instance
        """
        ...



"""
pre-built common_validators to perform common types of validation with high quality error messages. These should
also hook into pydantic's validation engine, specific implementations should inform how
"""
from spice_rack._common_validators import (
    _mutually_exclusive_fields as mutually_exclusive_fields,
)



from __future__ import annotations
from typing import final
from pathlib import Path
from fsspec.implementations.local import LocalFileSystem as FsSpecLocalFileSystem

from spice_rack._fs_ops._file_systems import _base
from spice_rack._fs_ops import _path_strs


__all__ = (
    "LocalFileSystem",
)


@final
class LocalFileSystem(_base.AbstractFileSystem, class_id="local"):
    """file system implementation for the local file system"""

    def build_fsspec_file_system(self) -> FsSpecLocalFileSystem:
        return FsSpecLocalFileSystem()

    @classmethod
    def get_fs_specific_prefix(cls) -> str:
        return "/"

    def get_home_dir(self) -> _path_strs.AbsoluteDirPathStr:
        return _path_strs.AbsoluteDirPathStr(str(Path().home()))



from spice_rack._fs_ops import (
    _fs_models as fs_models,
    _file_info as file_info,
    _file_systems as file_systems,
    _path_strs as path_strs,
    _exceptions as exceptions,
    _constraints as constraints
)

# maintain simplify imports
from spice_rack._fs_ops._fs_models import *  # ruff: noqa: SLF001
from spice_rack._fs_ops._file_info import *



from __future__ import annotations
import pathlib
from abc import abstractmethod
from devtools import pformat
import typing as t
import pydantic

from spice_rack import _bases, _logging
from spice_rack._fs_ops import (
    _path_strs,
    _file_systems,
)

if t.TYPE_CHECKING:
    from spice_rack._fs_ops._fs_models._file import FilePath
    from spice_rack._fs_ops._fs_models._dir import DirPath
    from spice_rack._fs_ops._fs_models._types import FileOrDirPathT


__all__ = (
    "AbstractFileSystemObj",
)


class AbstractFileSystemObj(
    _bases.DispatchableValueModelBase,
    _logging.log_extra.LoggableObjMixin
):
    """
    Base class for all file system objects in this package
    All file and dir objects inherit this class
    """
    path: _path_strs.FileOrDirAbsPathT = pydantic.Field(
        description="the path string on the underlying file system. Must be absolute. You can use the "
                    "file system-specific format, e.g. 'gs://bucket/file.txt' and we will convert it to "
                    "/bucket/file.txt' and make sure the file_system aligns"
    )
    file_system: _file_systems.AnyFileSystemT = pydantic.Field(
        description="the file system object for this file. if not specified, we'll infer it from the path.",
        default=None
    )

    @pydantic.model_validator(mode="before")
    def _handle_str(cls, data: t.Any) -> t.Any:
        if isinstance(data, (pathlib.Path, str)):
            data = cls.init_from_str(str(data)).model_dump()

        file_system_key = "file_system"
        path_key = "path"

        if isinstance(data, dict):
            raw_file_system = data.get(file_system_key)
            raw_path: str = data.get(path_key)
            if raw_path is None:
                raise ValueError(
                    f"no 'path' specified? data: \n{pformat(data)}"
                )
            else:
                raw_path = str(raw_path)
            file_system: _file_systems.AnyFileSystemT

            if raw_file_system is None:
                file_system = _file_systems.infer_file_system(raw_path)
            else:
                file_system = _file_systems.AnyFileSystemTypeAdapter.validate_python(
                    raw_file_system
                )
            data["file_system"] = file_system

            # either way we want to clean the path
            # either way we clean the path
            path_formatted = file_system.clean_raw_path_str(raw_path)
            data[file_system_key] = file_system
            data[path_key] = path_formatted

        return data

    @property
    def file_system_type(self) -> _bases.dispatchable.ClassId:
        """get the class id of the file system tied to this path object"""
        return self.file_system.get_class_id()

    @t.final
    def exists(self) -> bool:
        """
        returns True if this file system object exists, false otherwise
        """
        return self.file_system.exists(self.path)

    def __repr__(self) -> str:
        """
        deprecated, we want to remove all usage of this
        in logging and error messaging and use special_repr only
        """
        return self.special_repr()

    def __hash__(self) -> int:
        return hash(
            (self.path, self.file_system)
        )

    @abstractmethod
    def _special_repr_short(self) -> str:
        ...

    @abstractmethod
    def _special_repr_verbose(self) -> str:
        ...

    @t.final
    def special_repr(self, verbose: bool = False) -> str:
        if verbose:
            return self._special_repr_verbose()
        else:
            return self._special_repr_short()

    def __eq__(self, other: AbstractFileSystemObj) -> bool:
        if type(self) == type(other):  # noqa
            if hash(self) == hash(other):
                return True
        return False

    @abstractmethod
    def delete(
            self,
            if_non_existent: t.Literal["raise", "return"] = "return"
    ) -> None:
        ...

    @t.overload
    def build_like(self, path: _path_strs.AbsoluteFilePathStr) -> FilePath:
        ...

    @t.overload
    def build_like(self, path: _path_strs.AbsoluteDirPathStr) -> DirPath:
        ...

    @t.overload
    def build_like(self, path: str) -> FileOrDirPathT:
        ...

    def build_like(
            self, path: t.Union[str, _path_strs.AbsoluteFilePathStr, _path_strs.AbsoluteDirPathStr]
    ) -> FileOrDirPathT:
        """build a new file or dir path instance using the specified file path and the file
        system tied to this instance
        """
        from spice_rack._fs_ops._fs_models._file import FilePath
        from spice_rack._fs_ops._fs_models._dir import DirPath

        parsed_path = _path_strs.FileOrDirAbsPathTypeAdapter.validate_python(
            path
        )
        if isinstance(parsed_path, _path_strs.AbsoluteFilePathStr):
            return FilePath(
                path=parsed_path,
                file_system=self.file_system
            )
        elif isinstance(parsed_path, _path_strs.AbsoluteDirPathStr):
            return DirPath(
                path=parsed_path,
                file_system=self.file_system
            )
        else:
            raise ValueError(
                f"'{parsed_path}' is type {type(parsed_path)} which is unexpected"
            )

    def get_parent(self) -> DirPath:
        """return the dir path instance of the parent
        of the current dir or file path instance"""
        parent_path = self.path.get_parent()
        return self.build_like(path=parent_path)

    @abstractmethod
    def get_name(self) -> str:
        """get simple name, i.e. the most terminal chunk in the path"""
        ...

    def ensure_exists(self) -> None:
        """raise error if the path doesn't exist"""
        self.file_system.ensure_exists(self.path)

    def ensure_nonexistent(self) -> None:
        """raise an error if the path does exist"""
        self.file_system.ensure_nonexistent(self.path)

    def as_str(self) -> str:
        """
        return a contextualized str representation of the path,
        meaning the file system-specific prefix is used.
        """
        return self.file_system.contextualize_abs_path(self.path)

    def __get_logger_data__(self) -> t.Dict:
        return {
            "type": self.__class__.__name__,  # noqa
            "path": self.as_str(),
            "file_system_info": self.file_system.__get_logger_data__()
        }

    @classmethod
    @abstractmethod
    def init_from_str(cls, raw_str: str) -> AbstractFileSystemObj:
        ...



from __future__ import annotations
import typing as t
import pydantic

from spice_rack._ts_service._timestamp import Timestamp


__all__ = (
    "CreatedAtFieldAnn",
    "CreatedAtWithDefaultFieldAnn"
)


CreatedAtFieldAnn = t.Annotated[
    Timestamp,
    pydantic.Field(
        description="the timestamp when this object was created",
    )
]


CreatedAtWithDefaultFieldAnn = t.Annotated[
    Timestamp,
    pydantic.Field(
        description="the timestamp when this object was created, if not specified, "
                    "we'll create a timestamp at current time",
        default_factory=Timestamp.now
    )
]



from __future__ import annotations
import typing as t
from inspect import getdoc
from pydantic import Field

from spice_rack._bases import _value_model
from spice_rack._bases._exception._exception_base import CustomExceptionBase
from spice_rack._bases._exception._error_info import ErrorInfoBase, ErrorPayload
from spice_rack._bases._exception._external_wrapper import WrappedExternalException


__all__ = (
    "HttpException",
    "HttpErrorResponseInfo",
    "UnknownExceptionWrapper",
    "RequestValidationException",
)


class _UnknownExceptionErrorInfo(ErrorInfoBase):
    """
    An unexpected exception was encountered, i.e. not one of our subclasses.
    Wrapping it with this class allows us to work it into our exception handling and other
    tooling.
    """
    caught_exception: WrappedExternalException = Field(
        description="the exception we encountered"
    )


class UnknownExceptionWrapper(CustomExceptionBase[_UnknownExceptionErrorInfo]):
    """
    An unexpected exception was encountered, i.e. not one of our subclasses.
    Wrapping it with this class allows us to work it into our
    exception handling and other tooling.

    Notes:
        - a custom "detail" can be specified, or left blank, in which case we use the str
          representation of the caught exception
    """
    def __init__(
            self,
            caught_exception: Exception,
            detail: t.Optional[str] = None,
            verbose: bool = True,
            extra_info: t.Optional[dict] = None
    ):
        error_info = {
            "caught_exception": caught_exception,
        }
        if detail is None:
            detail = str(caught_exception)
        super().__init__(
            detail=detail,
            error_info=error_info,
            extra_info=extra_info,
            verbose=verbose
        )


class HttpErrorResponseInfo(_value_model.ValueModelBase):
    """
    Holds the status code and the exception class for a given endpoint.
    This is used to inform endpoints about the potential error responses, and their
    schemas.
    """
    status_code: int = Field(
        description="the status code for the response in question"
    )
    custom_desc: t.Optional[str] = Field(
        description="the custom specified description for this status code for the given endpoint"
    )
    exception_type: t.Type[CustomExceptionBase] = Field(
        description="the type of exception this error response will contain. This will "
                    "dictate schema of the resulting payload associated with the "
                    "response in question."
    )

    def get_desc(self) -> str:
        """
        use custom description if we have it, and use the docstring of the exception class if not
        """
        if self.custom_desc:
            return self.custom_desc
        else:
            docstring_maybe = getdoc(self.exception_type)
            if docstring_maybe:
                return docstring_maybe
            else:
                return f"a '{self.exception_type.__name__}' exception"


ErrorInfoTV = t.TypeVar("ErrorInfoTV", bound=ErrorInfoBase)


# todo: will delete one of these
class HttpErrorResponse(_value_model.ValueModelBase, t.Generic[ErrorInfoTV]):
    """
    generic container holding an exception payload and a specified status code.

    This is not an exception, so must be treated as a valid return type, and thus should be
    caught and the proper response raised.
    """
    # class Config:
    #     arbitrary_types_allowed = True

    status_code: int = Field(
        description="the status code for the response in question"
    )
    error_payload: ErrorPayload[ErrorInfoTV] = Field(
        description="the payload containing data about this error"
    )


class HttpException(Exception, t.Generic[ErrorInfoTV]):
    """
    generic container holding an exception instance and a specified status code.

    This is an exception
    """
    def __init__(
            self,
            status_code: int,
            error_payload: ErrorPayload[ErrorInfoTV]
    ):
        self.status_code = status_code
        self.error_payload = error_payload
        super().__init__(error_payload.detail)


# todo: figure out the correct inner type
_PydanticValidationErrorT: t.TypeAlias = t.Any
"""this is the data type of pydantic validation errors"""


class _RequestValidationErrorInfo(ErrorInfoBase):
    """
    The request fails to validate, so we don't even get the execute the endpoint.
    This is a wrapper around FastAPI's RequestValidationError.
    """
    issues: list[_PydanticValidationErrorT] = Field(
        description="the validation errors encountered"
    )


class RequestValidationException(
    CustomExceptionBase[_RequestValidationErrorInfo]
):
    """
    The request fails to validate, so we don't even get the execute the endpoint.
    This is a wrapper around FastAPI's RequestValidationError.
    """
    def __init__(
            self,
            detail: str,
            errors: list[_PydanticValidationErrorT],
            verbose: bool = True,
            extra_info: t.Optional[dict] = None
    ):
        error_info = {
            "issues": errors,
        }

        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=verbose,
            extra_info=extra_info,
        )



from __future__ import annotations
import os
from abc import abstractmethod
import typing as t
import pydantic

from spice_rack import _bases
from spice_rack._fs_ops import _path_strs

if t.TYPE_CHECKING:
    from spice_rack._fs_ops._fs_models._base import AbstractFileSystemObj
    from spice_rack._fs_ops._fs_models._file import FilePath
    from spice_rack._fs_ops._fs_models._dir import DirPath


__all__ = (
    "DeferredFilePath", "DeferredDirPath",
)


class _DeferredPath(_bases.DispatchableValueModelBase):
    env_var_key: str

    def _get_env_var_val(self) -> DirPath:
        from spice_rack._fs_ops._fs_models._dir import DirPath

        env_val_maybe = os.environ.get(self.env_var_key)
        if env_val_maybe is None:
            raise ValueError(
                f"'{self.env_var_key}' not found in the environment"
            )
        return DirPath.model_validate(env_val_maybe)

    @abstractmethod
    def evaluate(self) -> AbstractFileSystemObj:
        ...


class DeferredFilePath(_DeferredPath):
    """
    a file path, where part of the path is contained in an environment variable, that is
    looked up when we call 'evaluate' method to convert this instance to a FilePath instance.
    """
    rel_path: _path_strs.RelFilePathStr

    @pydantic.model_validator(mode="before")
    def _handle_str(cls, data: t.Any) -> t.Any:
        if isinstance(data, str):
            if "/" not in data:
                raise ValueError(
                    f"'{data}' cannot be parsed as "
                    f"'{cls.get_class_id()}' bc it has not relative file path beyond"
                    f"the deferred environment key"
                )

            split = data.split("/")
            env_var_key = split[0].replace("$", "")
            rel_path = "/".join(split[1:])

            data = {"env_var_key": env_var_key, "rel_path": rel_path}
        return data

    def evaluate(self) -> FilePath:
        env_val = self._get_env_var_val()
        return env_val.joinpath(self.rel_path)


class DeferredDirPath(_DeferredPath):
    """
    a dir path, where part or all of the path is contained in an environment variable, that is
    looked up when we call 'evaluate' method to convert this instance to a DirPath instance.
    """
    rel_path: t.Optional[_path_strs.RelDirPathStr]

    @pydantic.model_validator(mode="before")
    def _handle_str(cls, data: t.Any) -> t.Any:
        if isinstance(data, str):
            split = data.split("/")
            env_var_key = split[0].replace("$", "")
            rel_path = "/".join(split[1:])
            if not rel_path or rel_path == "/":
                rel_path = None
            data = {"env_var_key": env_var_key, "rel_path": rel_path}
        return data

    def evaluate(self) -> DirPath:
        env_val = self._get_env_var_val()
        if self.rel_path:
            return env_val.joinpath(self.rel_path)
        else:
            return env_val



from __future__ import annotations
from abc import ABC, abstractmethod
import typing as t

from spice_rack._logging._log_aug._container import ExtraLogData


__all__ = (
    "LoggableObjMixin",
)


class LoggableObjMixin(ABC):
    """
    provides interface for customizing how our logger formats a given class
    """

    # make sure this is the same as the method we specify in the container class
    @abstractmethod
    def __get_logger_data__(self) -> t.Union[str, t.Dict, ExtraLogData]:
        """
        mixin that informs which method to implement to add a hook
        that our logger will use to format how we represent this object
        in the logs.

        Returns: a dict of json-encodeable data
        """
        raise NotImplementedError()

    @t.final
    def __format_for_logger__(self) -> ExtraLogData:
        """this calls '__get_logger_data__' to get the data to log"""
        logger_data = self.__get_logger_data__()
        return ExtraLogData.parse_from_raw_data(data=logger_data)



from spice_rack._fs_ops._fs_models._base import *
from spice_rack._fs_ops._fs_models._file import *
from spice_rack._fs_ops._fs_models._dir import *
from spice_rack._fs_ops._fs_models._deferred import *
from spice_rack._fs_ops._fs_models._constraints import *
from spice_rack._fs_ops._fs_models._types import *

from spice_rack._fs_ops._fs_models import (
    _constraints as constraints,
    _types as types,
    _deferred as deferred
)



from __future__ import annotations
from typing import Sequence, TypeVar, Hashable


__all__ = (
    "get_difference_between_string_seqs",
    "get_difference_between_sets"
)


def get_difference_between_string_seqs(
        a: Sequence[str], b: Sequence[str]
) -> tuple[list[str], list[str]]:
    """
    compares 2 sequences of strings and returns the strings
    in A not in B and the strings in B not in A
    """
    a_set = set([str(aa) for aa in a])
    b_set = set([str(bb) for bb in b])

    in_a_not_b = [aa for aa in a_set if aa not in b]
    in_b_not_a = [bb for bb in b_set if bb not in a]
    return sorted(in_a_not_b), sorted(in_b_not_a)


SetItemTV = TypeVar("SetItemTV", bound=Hashable)


def get_difference_between_sets(
        a: set[SetItemTV],
        b: set[SetItemTV]
) -> tuple[set[SetItemTV], set[SetItemTV]]:
    total_items = a.union(b)

    items_in_a_not_b = set()
    items_in_b_not_a = set()

    for item in total_items:
        if item not in a:
            items_in_b_not_a.add(item)
        elif item not in b:
            items_in_a_not_b.add(item)
    return items_in_a_not_b, items_in_b_not_a



from __future__ import annotations
import datetime as dt
import pydantic

from spice_rack._bases._mixins._mixin_base import PydanticMixinBase


__all__ = ("CreatedAtMixin", "UpdatedAtMixin", "CreatedAtUpdatedAtMixin")


def _get_current_utc_datetime() -> dt.datetime:
    return dt.datetime.utcnow()


class CreatedAtMixin(PydanticMixinBase):
    """
    adds a datetime field, 'created_at' that can be timezone aware or naive
    """
    created_at: dt.datetime = pydantic.Field(
        description="timestamp the object was created at",
        default_factory=_get_current_utc_datetime
    )

    @pydantic.field_serializer("created_at", when_used="json-unless-none", check_fields=True)
    def _use_iso_for_created_at(self, data: dt.datetime):
        return data.isoformat()

    @property
    def created_at_str(self) -> str:
        return self.created_at.isoformat()


class UpdatedAtMixin(PydanticMixinBase):
    """
    adds a datetime field, 'updated_at' that can be timezone aware or naive
    """
    updated_at: dt.datetime = pydantic.Field(
        description="timestamp the object was last updated",
        default_factory=_get_current_utc_datetime
    )

    @pydantic.field_serializer("updated_at", when_used="json-unless-none", check_fields=True)
    def _use_iso_for_updated_at(self, data: dt.datetime):
        return data.isoformat()

    @property
    def updated_at_str(self) -> str:
        return self.updated_at.isoformat()


class CreatedAtUpdatedAtMixin(CreatedAtMixin, UpdatedAtMixin):
    """
    adds 2 datetime fields, 'created_at' and 'updated_at' that can be timezone aware or naive
    """
    ...



from __future__ import annotations
from typing import TYPE_CHECKING, Type, Optional
from pydantic import Field

from spice_rack import _bases

if TYPE_CHECKING:
    from spice_rack._frozen_registry._base import FrozenRegistryBase

__all__ = (
    "KeyAlreadyExistsException",
)


class _KeyAlreadyExistsErrorInfo(_bases.exceptions.ErrorInfoBase):
    """trying to add item but something with that key already exists"""
    registry_type: str = Field(description="the name of the registry class that made this error")
    duplicate_key: str = Field(description="the key that already exists in this registry")


class KeyAlreadyExistsException(
    _bases.exceptions.CustomExceptionBase[_KeyAlreadyExistsErrorInfo]
):
    """trying to add item but something with that key already exists"""
    def __init__(
            self,
            duplicate_key: str,
            registry_type: Type[FrozenRegistryBase],
            verbose: bool = True,
            extra_info: Optional[dict] = None
    ):
        detail = f"an item with the key, '{duplicate_key}', already exists"

        error_info = {
            "registry_type": registry_type.__name__,
            "duplicate_key": duplicate_key,
        }
        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=verbose,
            extra_info=extra_info
        )



from __future__ import annotations
import typing as t
import pydantic

from spice_rack._fs_ops import _path_strs, _file_systems
from spice_rack._fs_ops._fs_models._base import AbstractFileSystemObj

if t.TYPE_CHECKING:
    from spice_rack._fs_ops._fs_models._file import FilePath


__all__ = (
    "DirPath",
)


class DirPath(AbstractFileSystemObj):
    """
    this class represents a directory on a file system,
    with the absolute full path string, and the file system instance
    """
    path: _path_strs.AbsoluteDirPathStr = pydantic.Field(
        description="the path string on the underlying file system. Must be absolute. You can use the "
                    "file system-specific format, e.g. 'gs://bucket/dir/' and we will convert it to "
                    "/bucket/dir/' and make sure the file_system aligns"
    )

    def _special_repr_short(self) -> str:
        return f"Dir[path='{self.path}', file_system_type={self.file_system.special_repr()}]"

    def _special_repr_verbose(self) -> str:
        return f"Dir[path='{self.path}', file_system_type={self.file_system.special_repr()}]"

    def delete(
            self,
            if_non_existent: t.Literal["raise", "return"] = "return",
            recursive: bool = True,
    ) -> None:
        """
        delete the directory. recursively deleting all contents if recursive is true.
        """
        return self.file_system.delete_dir(
            self.path,
            recursive=recursive,
            if_non_existent=if_non_existent
        )

    def iter_dir(self) -> t.Iterator[t.Union[FilePath, DirPath]]:
        for path_i in self.file_system.iter_dir_contents(path=self.path):
            yield self.build_like(path_i)

    def iter_dir_contents_files_only(
            self,
            recursive: bool = True
    ) -> list[FilePath]:
        for path_i in self.file_system.iter_dir_contents_files_only(
                self.path,
                recursive=recursive
        ):
            yield self.build_like(path=path_i)

    @t.overload
    def joinpath(self, relative_path: _path_strs.RelFilePathStr) -> FilePath:
        ...

    @t.overload
    def joinpath(self, relative_path: _path_strs.RelDirPathStr) -> DirPath:
        ...

    @t.overload
    def joinpath(self, relative_path: str) -> t.Union[FilePath, DirPath]:
        ...

    def joinpath(
            self,
            relative_path: t.Union[str, _path_strs.FileOrDirRelPathT]
    ) -> t.Union[FilePath, DirPath]:
        """get a new FilePath or DirPath by combining this DirPath instance with a provided
        relative path.
        """
        rel_path_obj = _path_strs.FileOrDirRelPathTypeAdapter.validate_python(
            relative_path
        )
        new_path_any = self.path.joinpath(rel_path_obj)
        if isinstance(new_path_any, _path_strs.AbsoluteDirPathStr):
            new_path = new_path_any
            return DirPath(
                path=new_path,
                file_system=self.file_system
            )
        elif isinstance(new_path_any, _path_strs.AbsoluteFilePathStr):
            new_path = new_path_any
            from spice_rack._fs_ops._fs_models._file import FilePath

            return FilePath(
                path=new_path,
                file_system=self.file_system
            )
        else:
            raise ValueError(
                f"'{new_path_any}' is not valid type, type: {type(new_path_any)}"
            )

    def get_name(self) -> str:
        """get simple name, i.e. the most terminal chunk in the path"""
        return self.path.get_name()

    def make_self(
            self,
            if_exists: t.Literal["raise", "return"] = "return",
    ) -> None:
        """equivalent of mkdir"""
        self.file_system.make_dir(self.path, if_exists=if_exists)

    def download_locally(
            self,
            dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteDirPathStr:
        return self.file_system.download_dir_locally(
            self.path,
            dest_dir
        )

    @classmethod
    def init_from_str(cls, raw_str: str) -> DirPath:
        if raw_str.startswith("$"):
            from spice_rack._fs_ops._fs_models._deferred import DeferredDirPath
            return DeferredDirPath.model_validate(raw_str).evaluate()
        else:
            inferred_fs = _file_systems.infer_file_system(raw_str)
            path_str = inferred_fs.clean_raw_path_str(raw_str)
            dir_path = _path_strs.AbsoluteDirPathStr(path_str)
            return DirPath(path=dir_path, file_system=inferred_fs)



from __future__ import annotations
import typing as t
from abc import abstractmethod
from pathlib import Path

from spice_rack import _bases
from spice_rack._fs_ops import _exceptions


__all__ = (
    "AbstractPathStr",
)


class AbstractPathStr(_bases.special_str.SpecialStrBase):
    """Base class for all path string classes"""

    @classmethod
    def _parse_non_str(cls, root_data: t.Any) -> str:
        if isinstance(root_data, Path):
            str_val: str = str(root_data)

            try:
                if root_data.is_dir():
                    if not str_val.endswith("/"):
                        str_val += "/"

            # might have permissions issue with the path, but that is ok
            except Exception:  # noqa
                pass

            return str_val
        else:
            return super()._parse_non_str(root_data)

    @classmethod
    def _check_str_val(cls, __raw_str: str) -> t.List[str]:
        """checks str val for issues, returning a list of them if any found"""
        return []

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        issues = cls._check_str_val(root_data)
        if issues:
            raise _exceptions.InvalidPathStrException(
                detail=f"raw path str, '{root_data}', "
                       f"not valid for '{cls.get_cls_name()}' path type.",
                raw_str=root_data,
                issues=issues,
            )
        else:
            return root_data

    @classmethod
    def get_cls_name(cls) -> str:
        return cls.__name__

    @abstractmethod
    def get_name(self) -> str:
        """get simple name, i.e. the most terminal chunk in the path"""
        ...

    @abstractmethod
    def get_parent(self) -> AbstractPathStr:
        ...



from __future__ import annotations
import typing as t
from pydantic import PrivateAttr
from pydantic.v1.fields import DeferredType

from spice_rack import _bases, _utils
from spice_rack._frozen_registry import _exceptions

__all__ = (
    "FrozenRegistryBase",
    "FrozenRegistryOutBase",
)


# Note: the "_RegistryItemTV" class must have the specified key_attr_name attribute,
# but we can't specify this via base class, so we'll just manually check for it here.
_RegistryItemTV = t.TypeVar("_RegistryItemTV", bound=_bases.ValueModelBase)
_RegistryItemKeyTV = t.TypeVar("_RegistryItemKeyTV", bound=str)
Self = t.TypeVar("Self", bound="FrozenRegistryBase")


class FrozenRegistryBase(
    _bases.RootModel[tuple[_RegistryItemTV, ...]],
    t.Generic[_RegistryItemTV, _RegistryItemKeyTV]
):
    """
    Base class for a registry of items, where we can control how we get and add things to it.
    We use this pattern in various places, so consolidating behavior to a base class here
    minimizes the small code drift between the different implementations.

    Notes:
        - Specify the subclass kwarg 'is_base' as True to delay the item class validation
          otherwise performed in '__init_subclass__'

    """
    _distinct_keys: set[_RegistryItemKeyTV] = PrivateAttr(default=None)

    def __init_subclass__(
            cls,
            is_base: bool = False,
            **kwargs
    ) -> None:
        super().__init_subclass__(**kwargs)
        if not is_base:
            _validate_item_cls(cls)
        return

    @classmethod
    def get_key_attr_name(cls) -> str:
        """overwrite this to customize the attr we use as the key for each item"""
        return "key"

    def _post_init_setup(self) -> None:
        _distinct_keys = set(
            self._get_item_key_val(item) for item in self.iter_items()
        )
        self._distinct_keys = _distinct_keys  # noqa

    def _post_init_validation(self) -> None:
        # validate the item class post init, rather than init subclass
        _validate_item_cls(self.__class__)
        assert self._distinct_keys is not None, \
            f"'{self.get_cls_name()}' didn't set _distinct_keys private attr"

        # ensure no dupe keys
        key_set: set[_RegistryItemKeyTV] = set()
        dupe_keys: dict[_RegistryItemKeyTV, int] = {}
        for item in self.iter_items():
            item_key = self._get_item_key_val(item)
            if item_key in key_set:
                # must be at least 1
                num_appearances = dupe_keys.get(item_key, 1)
                dupe_keys[item_key] = num_appearances + 1
            else:
                key_set.add(item_key)
        if dupe_keys:
            raise _exceptions.DuplicateKeysInRegistryException(
                registry_type=type(self),
                duplicate_keys=dupe_keys,
                extra_info={}
            )

    @classmethod
    def _get_default_value(cls) -> tuple[_RegistryItemTV, ...]:
        return ()

    @classmethod
    def init_empty(cls: t.Type[Self]) -> Self:
        return cls(root=())

    @classmethod
    def get_item_cls(cls) -> t.Type[_RegistryItemTV]:
        item_ann = t.get_args(cls.model_fields["root"].annotation)[0]
        if isinstance(item_ann, t.TypeVar):
            raise ValueError(f"'{cls.get_cls_name()}' hasn't set their item type var yet")
        else:
            return item_ann

    @classmethod
    def get_key_cls(cls) -> t.Type[_RegistryItemKeyTV]:
        """default behavior is str, but we can customize this by overwriting this class-method"""
        return str

    def _get_item_key_val(self, item: _RegistryItemTV) -> _RegistryItemKeyTV:
        key_val = getattr(item, self.get_key_attr_name())
        return _utils.check_type(key_val, self.get_key_cls())

    @classmethod
    def _items_equal(cls, item1: _RegistryItemTV, item2: _RegistryItemTV) -> bool:
        """
        default behavior is direct equality, overwrite this to customize behavior of the
        registry __eq__ dunder
        """
        return item1 == item2

    def size(self) -> int:
        return len(self.root)

    def iter_items(self: Self) -> t.Iterator[_RegistryItemTV]:
        """iterator over all items"""
        for obj in self.root:
            yield obj

    def __len__(self) -> int:
        return len(self.root)

    def __getitem__(self, ix: int) -> _RegistryItemTV:
        try:
            return self.root[ix]
        except Exception as e:
            raise _exceptions.ItemIndexInvalidException(
                invalid_ix=ix,
                registry_type=type(self),
                registry_size=self.size(),
                caught_exception=e,
                verbose=True,
                extra_info={}
            )

    def __iter__(self: Self) -> t.Iterator[_RegistryItemTV]:
        for item in self.iter_items():
            yield item

    def list_keys(self: Self) -> list[_RegistryItemKeyTV]:
        """
        Returns: list all keys in the registry
        """
        return list(self._distinct_keys)

    def get_item_maybe(self: Self, __key: _RegistryItemKeyTV) -> t.Optional[_RegistryItemTV]:
        """
        get an item for the key or return None if nothing is found

        Args:
            __key: the key we're looking for

        Returns: an instance of _RegistryItemTV or None
        """
        for obj in self.iter_items():
            if self._get_item_key_val(obj) == __key:
                return obj
        return

    def get_item(self, __key: _RegistryItemKeyTV) -> _RegistryItemTV:
        """
        get an item for the key, or raise an exception

        Args:
            __key: the key we're looking for

        Returns: an instance of _RegistryItemTV

        Raises:
            KeyNotFoundException: if nothing is found
        """
        item_maybe = self.get_item_maybe(__key)
        if item_maybe is None:
            raise _exceptions.KeyNotFoundException(
                missing_key=__key,
                registry_type=type(self),
                keys_in_registry=self.list_keys(),
                extra_info={}
            )
        else:
            return item_maybe

    def key_exists(self, __key: _RegistryItemKeyTV) -> bool:
        return __key in self._distinct_keys

    def raise_if_found(self, __key: _RegistryItemKeyTV) -> None:
        """
        convenience method that raises an exception if the key is found

        Raises:
            KeyAlreadyExistsException: if key is found
        """
        if __key in self._distinct_keys:
            raise _exceptions.KeyAlreadyExistsException(
                duplicate_key=__key,
                registry_type=type(self),
                extra_info={}
            )

    def raise_if_not_found(self, __key: _RegistryItemKeyTV) -> None:
        """
        convenience method that raises an exception if the key is not found

        Raises:
            KeyNotFoundException: if key is found
        """
        if __key not in self._distinct_keys:
            raise _exceptions.KeyNotFoundException(
                missing_key=__key,
                registry_type=type(self),
                keys_in_registry=self.list_keys(),
                extra_info={}
            )

    def with_new_items(
            self: Self,
            new_items: list[_RegistryItemTV],
            if_exists: t.Literal["raise", "replace", "skip"] = "replace"
    ) -> Self:
        """
        create a new instance of this registry with the new items, raising an exception if
        an item already exists with the same key and 'if_exists' equals 'raise'

        Args:
            new_items: the list of items we are adding
            if_exists: if 'raise' we raise an exception if an item already exists with this key,
                and if 'replace', we replace the existing item with the new item,
                if "skip", we don't add the new item, but don't raise an error either

        Returns: a new instance of this class

        Raises:
            KeyAlreadyExistsException: an item with the same key as the new item already exists
                and 'if_exists' is set to 'raise'
        """

        # remake the key set as we add each item, so duplicates within the group of new
        # items don't make their way in.
        item_dict: dict[_RegistryItemKeyTV, _RegistryItemTV] = {
            self._get_item_key_val(item): item for item in self.iter_items()
        }
        for new_item_i in new_items:
            new_item_i_key = self._get_item_key_val(new_item_i)

            if new_item_i_key in item_dict:
                if if_exists == "raise":
                    raise _exceptions.KeyAlreadyExistsException(
                        duplicate_key=new_item_i_key,
                        registry_type=type(self),
                        extra_info={}
                    )
                elif if_exists == "replace":
                    item_dict[new_item_i_key] = new_item_i

                elif if_exists == "skip":
                    continue

                else:
                    raise ValueError(f"unexpected value for 'if_exists': '{if_exists}'")
            else:
                item_dict[new_item_i_key] = new_item_i

        return t.cast(Self, self.__class__(tuple(item_dict.values())))

    def without_items(
            self: Self,
            keys: list[_RegistryItemKeyTV],
            if_not_found: t.Literal["raise", "continue"] = "continue"
    ) -> Self:
        """
        create a new registry, without the item tied to the key specified

        Args:
            keys: the list of keys to the items we are deleting
            if_not_found: if 'raise', we raise an error if nothing exists with a key,
                and if 'continue' we just move forward as if we deleted it.

        Returns: the new instance of Self

        Raises:
            KeyNotFoundException: if nothing is found for a key, and 'if_not_found' is 'raise'
        """

        # make a set of keys to check against
        keys_to_delete: set[_RegistryItemKeyTV] = set()
        for key_i in keys:
            if key_i not in self._distinct_keys:
                if if_not_found == "raise":
                    # will raise for us
                    self.raise_if_not_found(key_i)
                else:
                    continue
            else:
                keys_to_delete.add(key_i)

        updated_items: list[_RegistryItemTV] = []
        for item in self.iter_items():
            if self._get_item_key_val(item) in keys_to_delete:
                continue
            else:
                updated_items.append(item)
        res = type(self)(tuple(updated_items))
        return t.cast(Self, res)

    def with_new_item(
            self: Self,
            new_item: _RegistryItemTV,
            if_exists: t.Literal["raise", "replace", "skip"] = "replace"
    ) -> Self:
        """
        create a new instance of this registry with a new item, raising an exception if
        an item already exists with the same key and 'if_exists' equals 'raise'

        Args:
            new_item: the item we are adding
            if_exists: if 'raise' we raise an exception if an item already exists with this key,
                and if 'replace', we replace the existing item with the new item,
                if "skip", we don't add the new item, but don't raise an error either

        Returns: a new instance of this class

        Raises:
            KeyAlreadyExistsException: an item with the same key as the new item already exists
                and 'if_exists' is set to 'raise'
        """
        return self.with_new_items(new_items=[new_item], if_exists=if_exists)

    def without_item(
            self: Self,
            key: _RegistryItemKeyTV,
            if_not_found: t.Literal["raise", "continue"] = "continue"
    ) -> Self:
        """
        create a new registry, without the item tied to the key specified

        Args:
            key: key to the item we are deleting
            if_not_found: if 'raise', we raise an error if nothing exists with this key,
                and if 'continue' we just move forward as if we deleted it.

        Returns: the new instance of Self

        Raises:
            KeyNotFoundException: if nothing is found for the key, and 'if_not_found' is 'raise'
        """
        return self.without_items(keys=[key], if_not_found=if_not_found)

    def __eq__(self: Self, other: t.Any) -> bool:
        if isinstance(other, type(self)):

            # if different size, we know not equal
            if self.size() != other.size():
                return False

            # check same keys
            self_keys = set(self.list_keys())
            other_keys = set(other.list_keys())

            # extra and missing from the POV of self,
            #   i.e. extra_keys = keys in "self" not in "other"
            extra_keys, missing_keys = _utils.get_difference_between_sets(
                self_keys, other_keys
            )
            if extra_keys or missing_keys:
                return False

            # check each object individually
            for key in self_keys:
                self_obj = self.get_item(key)
                other_obj = other.get_item(key)
                if not self._items_equal(self_obj, other_obj):
                    return False

            # if we make it here, they are equal
            return True
        else:

            # not same type, so definitely False
            return False

    def special_repr(self: Self) -> str:
        return (f"{self.__class__.__name__}"
                f"[key_field='{self.get_key_attr_name()}', size={self.size()}]")


def _validate_item_cls(_registry_cls: t.Type[FrozenRegistryBase]) -> None:
    """
    Executes some manual checks that the registry's item class meets the requirements expected by
    the frozen registry container
    """
    # this means it is the intermediate/passthrough type made when we set up a class with a
    # parameterized base. We skip this validation in this case
    if _registry_cls.get_cls_name().startswith("FrozenRegistryBase["):
        return

    # this is a concrete frozen registry class, i.e. one we plan to instantiate and use.

    # Check the item type isn't a generic, i.e. the type var has been replaced with a proper class.
    item_cls = _registry_cls.get_item_cls()

    # todo: this DeferredType check might be unnecessary in pydantic v2.
    if isinstance(item_cls, DeferredType):
        raise ValueError(
            f"'{_registry_cls.get_cls_name()}' has a deferred type for the item class still"
        )

    elif isinstance(item_cls, t.TypeVar):
        raise ValueError(
            f"'{_registry_cls.get_cls_name()}' has a type var as the item type still, {item_cls}"
        )
    else:

        # make sure the object has a 'key' attribute with the correct annotation
        # this must be either a pydantic field or a property-decorated method
        item_cls = _utils.check_subclass(item_cls, _bases.ValueModelBase)
        key_attr_name = _registry_cls.get_key_attr_name()
        key_field_cls = _registry_cls.get_key_cls()

    # check if pydantic field
        if key_attr_name in item_cls.model_fields:
            key_field = item_cls.model_fields[key_attr_name]
            key_field_type_ann = key_field.annotation
            if key_field_type_ann != key_field_cls:
                raise ValueError(
                    f"the key field type ann must equal the key field class"
                    f"specified in this registry. "
                    f"We encountered {key_field_type_ann} and {key_field_cls}"
                )

        # check if defined as a property
        elif hasattr(item_cls, key_attr_name):

            # could be any type of class attribute
            key_method = getattr(item_cls, key_attr_name)

            # todo: could also add support for a no-arg method, but for now just tell user to make it a property
            # make sure a property-decorated function, with the correct return annotation
            if isinstance(key_method, property):
                getter_func: t.Callable = key_method.fget
                _getter_anns: t.Dict = t.get_type_hints(getter_func)
                if "return" not in _getter_anns:
                    raise ValueError(
                        f"You must specify a return annotation for the '{key_attr_name}' "
                        f"property on the '{item_cls.get_cls_name()}'to use it as a key"
                    )

                return_ann = _getter_anns["return"]
                if return_ann != key_field_cls:
                    raise ValueError(
                        f"'{key_attr_name}' on the '{item_cls.get_cls_name()}' maps to a property "
                        f"with a return annotation of '{return_ann}', and the '{_registry_cls.get_cls_name()}' "
                        f"registry class expects the key field to be type {key_field_cls}."
                    )

            # not a property
            else:
                raise ValueError(
                    f"the specified key attribute, '{key_attr_name}', must be either a pydantic field or a property,"
                    f" and the '{item_cls.get_cls_name()}' class defined the '{key_attr_name}' "
                    f"attribute as type {type(key_method)}"
                )

        # not found at all
        else:
            raise ValueError(
                f"'{item_cls.get_cls_name()}' doesn't have a '{key_attr_name}' field or property specified"
            )

        return None


class FrozenRegistryOutBase(_bases.RootModel[t.List[_RegistryItemTV]], t.Generic[_RegistryItemTV]):
    """base class for an api response model for a frozen registry"""
    ...



from __future__ import annotations
import typing as t
import pydantic

from spice_rack import _bases

if t.TYPE_CHECKING:
    from spice_rack._fs_ops import _path_strs, _file_systems, _file_info

__all__ = (
    "InvalidPathStrException",
    "NonExistentPathException",
    "PathAlreadyExistsException",
    "InvalidFileExtensionException",
    "InvalidFileMimeTypeException"
)

ErrorInfoBase = _bases.exceptions.ErrorInfoBase
CustomExceptionBase = _bases.exceptions.CustomExceptionBase


class _InvalidPathStrErrorInfo(ErrorInfoBase):
    """this is raised when one of the path string classes fails to validate a raw string"""
    raw_str: str
    issues: t.List[str]


class InvalidPathStrException(CustomExceptionBase[_InvalidPathStrErrorInfo]):
    """this is raised when one of the path string classes fails to validate a raw string"""
    def __init__(
            self,
            detail: str,
            raw_str: str,
            issues: t.Union[str, list[str]],
            verbose: bool = True,
            extra_info: t.Dict = None
    ):

        issue_lst: t.List[str]
        if isinstance(issues, str):
            issue_lst = [issues]
        else:
            issue_lst = issues

        super().__init__(
            detail=detail,
            error_info={
                "raw_str": raw_str,
                "issues": issue_lst
            },
            verbose=verbose,
            extra_info=extra_info
        )


class _NonExistentPathErrorInfo(ErrorInfoBase):
    """path not found on file system"""
    file_system_info: t.Dict = pydantic.Field(description="the file system we were using")
    path: str


class NonExistentPathException(CustomExceptionBase[_NonExistentPathErrorInfo]):
    def __init__(
            self,
            file_system: _file_systems.AbstractFileSystem,
            path: _path_strs.FileOrDirAbsPathT,
            verbose: bool = True,
            extra_info: t.Dict = None
    ):
        detail = "either the file path does not exist or we do not have access to it"
        error_info = {
            "file_system_info": file_system.__get_logger_data__(),
            "path": file_system.contextualize_abs_path(path)
        }
        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=verbose,
            extra_info=extra_info
        )


class _PathAlreadyExistsErrorInfo(ErrorInfoBase):
    """path found when we didn't expect to"""
    file_system_info: t.Dict = pydantic.Field(description="the file system we were using")
    path: str


class PathAlreadyExistsException(CustomExceptionBase[_PathAlreadyExistsErrorInfo]):
    def __init__(
            self,
            file_system: _file_systems.AbstractFileSystem,
            path: _path_strs.FileOrDirAbsPathT,
            verbose: bool = True,
            extra_info: t.Dict = None
    ):
        detail = "this path already exists"
        error_info = {
            "file_system_info": file_system.__get_logger_data__(),
            "path": file_system.contextualize_abs_path(path)
        }
        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=verbose,
            extra_info=extra_info
        )


class _InvalidFileExtensionErrorInfo(ErrorInfoBase):
    full_path: str
    file_ext_found: t.Optional[str]
    file_ext_choices: t.List[str]


class InvalidFileExtensionException(CustomExceptionBase[_InvalidFileExtensionErrorInfo]):
    def __init__(
            self,
            path: t.Union[_path_strs.RelFilePathStr, _path_strs.AbsoluteFilePathStr],
            file_ext_found: t.Optional[_file_info.FileExt],
            file_ext_choices: t.List[_file_info.FileExt],
            verbose: bool = True,
            extra_info: t.Dict = None,
    ):
        error_info = {
            "full_path": path,
            "file_ext_found": file_ext_found,
            "file_ext_choices": file_ext_choices,
        }
        if not file_ext_found:
            detail = (f"the path, '{path}' had no file extension, "
                      f"but one the specified file extensions was expected")
        else:
            detail = (f"the path, '{path}' had file extension, '{file_ext_found}' "
                      f"which is not one of the specified file extensions")
        super().__init__(
            detail=detail,
            error_info=error_info,
            extra_info=extra_info,
            verbose=verbose
        )


class _InvalidFileMimeTypeErrorInfo(ErrorInfoBase):
    full_path: str
    mime_type_found: t.Optional[str]
    mime_type_choices: t.List[str]


class InvalidFileMimeTypeException(CustomExceptionBase[_InvalidFileMimeTypeErrorInfo]):
    def __init__(
            self,
            path: _path_strs.FileOrDirAbsPathT,
            mime_type_found: t.Optional[str],
            mime_type_choices: t.List[str],
            verbose: bool = True,
            extra_info: t.Dict = None,
    ):
        error_info = {
            "full_path": path,
            "mime_type_found": mime_type_found,
            "mime_type_choices": mime_type_choices,
        }
        if not mime_type_found:
            detail = (f"the path, '{path}' had no mime type, "
                      f"but one the specified mime types was expected")
        else:
            detail = (f"the path, '{path}' had mime type, '{mime_type_found}' "
                      f"which is not one of the specified mime types")
        super().__init__(
            detail=detail,
            error_info=error_info,
            extra_info=extra_info,
            verbose=verbose
        )



from spice_rack._fs_ops._file_systems._base import *
from spice_rack._fs_ops._file_systems._local import *
from spice_rack._fs_ops._file_systems._gcs import *
from spice_rack._fs_ops._file_systems._sftp import *
from spice_rack._fs_ops._file_systems._sftp import *
from spice_rack._fs_ops._file_systems._fs_inference import *

AnyFileSystemT = AbstractFileSystem.build_dispatched_ann()
AnyFileSystemTypeAdapter = AbstractFileSystem.build_dispatcher_type_adapter()



from __future__ import annotations
import typing as t
import logging
from loguru import logger as loguru_logger
import warnings

from spice_rack._logging import _sinks


__all__ = (
    "configure_sink_for_service",
    "configure_logging_for_python_worker",
    "configure_logging_for_uvicorn_server",
)


def configure_sink_for_service(
        sink: _sinks.AbstractLogSink,
        custom_loguru_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
) -> None:
    custom_loguru_kwargs = {} if not custom_loguru_kwargs else custom_loguru_kwargs
    try:
        sink.setup(logger=loguru_logger, **custom_loguru_kwargs)

    # if the setup method raised a SinkSetupError on its own, we just directly raise that
    # except SinkSetupError as e:
    #     raise e

    # we convert other exceptions to SinkSetupError's
    except Exception as e:
        raise e
        # raise SinkSetupError(sink=sink, loguru_kwargs=loguru_kwargs, error=e)


class _UvicornInterceptHandler(logging.Handler):
    """allows loguru to work well with uvicorn servers"""
    def emit(self, record):
        # Get corresponding Loguru level if it exists
        from spice_rack._logging._logger_getter import get_logger
        try:
            level = get_logger().loguru_logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        get_logger().loguru_logger.opt(
            depth=depth, exception=record.exc_info
        ).log(level, record.getMessage())


def configure_logging_for_uvicorn_server(
        sinks: list[_sinks.AbstractLogSink],
) -> None:
    try:
        loguru_logger.remove(0)

    except ValueError as e:
        warnings.warn(str(e))

    except Exception as e:
        raise e

    # intercept everything at the root logger
    logging.root.handlers = [_UvicornInterceptHandler()]

    # changed from the tutorial
    # set to the lowest level and loguru-configured sinks will filter to their level
    logging.root.setLevel(logging.DEBUG)

    # remove every other logger's handlers
    # and propagate to root logger
    for name in logging.root.manager.loggerDict.keys():
        logging.getLogger(name).handlers = []
        logging.getLogger(name).propagate = True

    # configure loguru
    # this line was his, we configure from our setting object
    # get_logger().configure(handlers=[{"sink": sys.stdout, "serialize": JSON_LOGS}])
    for sink in sinks:
        configure_sink_for_service(sink=sink)


def configure_logging_for_python_worker(
        sinks: list[_sinks.AbstractLogSink],
) -> None:
    try:
        loguru_logger.remove(0)

    except ValueError as e:
        warnings.warn(str(e))

    except Exception as e:
        raise e

    for sink in sinks:
        configure_sink_for_service(sink=sink)



from __future__ import annotations
import typing as t
import pydantic


__all__ = (
    "RootModel",
)


RootTV = t.TypeVar("RootTV", )
SelfTV = t.TypeVar("SelfTV", bound="RootModel")


class RootModel(pydantic.RootModel[RootTV], t.Generic[RootTV]):
    """subclass pydantic's RootModel"""

    @classmethod
    def _get_default_value(cls) -> RootTV:
        raise ValueError(
            f"'{cls.__name__}' doesn't have a '_get_default_value' implemented so you must "
            f"specify the value, there is no default"
        )

    @pydantic.model_validator(mode="before")
    def _use_default(cls, data: t.Any) -> t.Any:
        if data is pydantic.fields.PydanticUndefined:
            return cls._get_default_value()
        else:
            return data

    @classmethod
    def get_cls_name(cls) -> str:
        return cls.__name__

    def _post_init_setup(self) -> None:
        """
        Overwrite this hook to set perform misc. logic before calling _post_init_validation.
        Common use case for this is initializing private attributes.
        Notes: Make sure to call super()._post_init_setup
        """
        return

    def _post_init_validation(self) -> None:
        """
        Overwrite this hook to set perform misc. validation logic on the instance, after
        all other common_validators have been executed.
        This is also called after the '_post_init_setup' hook is called
        Notes: Make sure to call super()._post_init_validation
        """
        return

    @pydantic.model_validator(mode="after")
    def _pydantic_post_init_val_hook(self: SelfTV) -> SelfTV:
        """
        Pydantic's hook that gets executed after we initialize an instance.
        rather than overwrite this, overwrite the '_post_init_setup' and '_post_init_validation'
        hooks
        """
        self._post_init_setup()
        self._post_init_validation()
        return self

    def json_dict(
            self,
            use_str_fallback: bool = True,
            **pydantic_kwargs
    ) -> t.Any:
        """
        convert into a json-encodeable data structure. whatever is encodeable version
        of RootTV, so if a list[str], this will return list of strings,
        if a pydantic model, this will return a dict.
        """
        pydantic_kwargs = pydantic_kwargs if pydantic_kwargs is not None else {}
        if use_str_fallback:
            fallback = str
        else:
            fallback = None
        return self.__pydantic_serializer__.to_python(
            self,
            mode="json",
            fallback=fallback,
            **pydantic_kwargs,
        )

    @classmethod
    def _import_forward_refs(cls) -> dict:
        """
        hook provided for convenience when you define a subclass with forward refs,
        to avoid circular import issue. Overwrite this method to do the imports and
        add the newly imported objects to a dict.

        Notes:
            - If you're not sure what to import and what keys to use, look at the stuff
              in the `if TYPE_CHECKING:` block, paste those exact imports in here and
              set them equal to their variable name within your current module's context.

        Returns:
            dict: a dict of references to update

        """
        return {}

    @classmethod
    def model_rebuild(
            cls,
            *,
            force: bool = False,
            raise_errors: bool = True,
            _parent_namespace_depth: int = 2,
            _types_namespace: dict[str, t.Any] | None = None,
    ) -> t.Optional[bool]:
        """
        This extends pydantic's builtin hook for updating forward refs.
        We call our special '_import_forward_refs' hook to automatically bring in the things
        imported there, but also if you want to use it like pydantic, i.e. call it directly and
        provide your refs to update via kwargs, that is still supported.

        Returns:
            None: the schema is already "complete" and rebuilding was not required.
            True: rebuilding was required and was successful
            False: rebuilding was required and not successful
        """
        _types_namespace_dict: dict[str, t.Any] = _types_namespace if _types_namespace else {}
        imported_refs = cls._import_forward_refs()
        _types_namespace_dict.update(imported_refs)
        return super().model_rebuild(
            force=force,
            raise_errors=raise_errors,
            _types_namespace=_types_namespace_dict,
            _parent_namespace_depth=_parent_namespace_depth
        )  # type: ignore



from __future__ import annotations
import typing as t
import typing_extensions
import uvicorn
from uvicorn.supervisors import Multiprocess, ChangeReload
from uvicorn.supervisors.basereload import BaseReload
from pydantic import Field
from spice_rack import _bases


__all__ = (
    "UvicornStartupConfig",
    "start_uvicorn"
)


class _UvicornSslConfig(_bases.ValueModelBase):
    """nested config for configuring ssl to uvicorn the server"""
    keyfile_path: str
    keyfile_password: str
    certfile_path: str


class UvicornStartupConfig(_bases.SettingsBase):
    """
    config subsection for starting up an uvicorn process, useful for APIs such as litestar or fastapi.

    default env_prefix is 'uvicorn_startup__'
    """
    host: str = Field(default="localhost", description="host name of the app process")
    port: int = Field(default=8000, description="port the app is listening on")
    root_path: str = Field(description="used in the fastapi app instance", default="")
    reload_mode: bool = Field(
        default=False, description="whether or not to auto-reload if any changes "
                                   "to the underlying folders occur."
                                   " This should be false in prod"
    )
    num_workers: int = Field(
        description="num server workers, we only support 1 at the moment.",
        default=1,
        gt=0,
    )

    proxy_headers: bool = Field(
        description="passed to uvicorn config",
        default=False
    )
    forwarded_allow_ips: t.Optional[t.Union[list[str], str]] = Field(
        description="passed to uvicorn config",
        default=None
    )
    ssl: t.Optional[_UvicornSslConfig] = Field(
        description="ssl config stuff for the user, don't use this in production",
        default=None
    )

    @classmethod
    def _get_env_prefix(cls) -> t.Optional[str]:
        return f"uvicorn_startup{cls._get_env_nested_delimiter()}"


def start_uvicorn(
        app: t.Any,
        config: UvicornStartupConfig,
) -> typing_extensions.Never:
    """
    serve an app using the uvicorn server configured with a specified config instance

    Args:
        app: the app we are serving. Should be something like FastApi or litestar instance
        config: the uvicorn config

    Returns:
        Never returns, runs until something kills the process
    """
    ssl_keyfile: t.Optional[str] = None
    ssl_keyfile_password: t.Optional[str] = None
    ssl_certfile: t.Optional[str] = None
    if config.ssl:
        ssl_keyfile = config.ssl.keyfile_path
        ssl_keyfile_password = config.ssl.keyfile_password
        ssl_certfile = config.ssl.certfile_path

    uvicorn_config = uvicorn.Config(
        app=app,
        host=config.host,
        port=config.port,
        reload=config.reload_mode,
        workers=config.num_workers,
        root_path=config.root_path,
        proxy_headers=config.proxy_headers,
        forwarded_allow_ips=config.forwarded_allow_ips,
        ssl_keyfile=ssl_keyfile,
        ssl_keyfile_password=ssl_keyfile_password,
        ssl_certfile=ssl_certfile
    )

    server = uvicorn.Server(config=uvicorn_config)
    supervisor_type: t.Optional[t.Union[t.Type[BaseReload], t.Type[Multiprocess]]] = None
    if uvicorn_config.should_reload:
        supervisor_type = ChangeReload
    if uvicorn_config.workers > 1:
        supervisor_type = Multiprocess
    if supervisor_type:
        sock = uvicorn_config.bind_socket()
        supervisor = supervisor_type(uvicorn_config, target=server.run, sockets=[sock])
        supervisor.run()
    else:
        server.run()

    raise Exception("Should have started a server before reaching this point")



from __future__ import annotations
import typing as t
# from http import HTTPStatus
import pydantic

from spice_rack._bases._value_model import ValueModelBase


__all__ = (
    "WrappedExternalException",
)


class WrappedExternalException(ValueModelBase):
    type: str = pydantic.Field(description="the name of the type of the error")
    detail: t.Union[str, dict] = pydantic.Field(
        description="The string representation of the error"
    )

    @pydantic.model_validator(mode="before")
    def _parse_exception(cls, data: t.Any) -> t.Any:
        if isinstance(data, Exception):
            # still get nice formatting if it is our internal error class
            from spice_rack._bases._exception._exception_base import CustomExceptionBase
            if isinstance(data, CustomExceptionBase):
                data = {
                    "type": type(data).__name__,
                    "detail": data.get_error_payload_inst().json_dict()
                }

            else:
                data = {
                    "type": type(data).__name__,
                    "detail": str(data)
                }

        return data



from spice_rack._gcp_auth._auth_strategies._base import *
from spice_rack._gcp_auth._auth_strategies._anon import *
from spice_rack._gcp_auth._auth_strategies._default import *
from spice_rack._gcp_auth._auth_strategies._service_acct import *

AnyGcpAuthT = _base.AbstractGcpAuthStrategy.build_dispatched_ann()



from __future__ import annotations
import typing as t

from spice_rack._logging._logger import Logger


__all__ = (
    "get_logger",
)


def get_logger(
        specified_service_name: t.Optional[str] = None,
) -> Logger:
    """
    function that returns a logger.
    This is here for backward compatibility of the api but you should be using
    Logger.get_logger instead
    """
    return Logger.get_logger(specified_service_name)



from __future__ import annotations
import typing as t
import datetime as dt
import pydantic
from functools import total_ordering
import zoneinfo

from spice_rack import _logging, _bases
from spice_rack._ts_service._tz_key import TimeZoneKey


__all__ = (
    "Timestamp",
)

_TzKeyT = t.Union[str, TimeZoneKey, t.Literal["local"]]
_PythonTimestampT = float  # seconds from epoch with decimals as sub-second measurements


TimestampSelfTV = t.TypeVar("TimestampSelfTV", bound="Timestamp")


@total_ordering
class Timestamp(_bases.ValueModelBase, _logging.log_extra.LoggableObjMixin):
    """
    This class represents timestamp, with microsecond precision, and the timezone info related to this timestamp
    and encapsulates common timestamp operations

    This class will automatically validate dt.datetime, dt.date and iso formatted strings. See extensions for
    more flexible parsing logic.
    """
    microseconds: int = pydantic.Field(
        description="microseconds from epoch, standardized so the timezone does not impact this value"
    )
    tz: TimeZoneKey = pydantic.Field(
        description="the timezone info for this timestamp, default is UTC",
    )

    @property
    def seconds(self) -> float:
        return self.microseconds / 1e6

    def to_python_timestamp(self) -> _PythonTimestampT:
        # convert to seconds from epoch, posix timestamp
        return float(self.microseconds) / 1e6

    def with_tz(self, __new_tz: t.Union[str, TimeZoneKey]) -> "Timestamp":
        """convert this Timestamp instance to one with the different timezone"""
        return Timestamp(
            microseconds=self.microseconds,
            tz=TimeZoneKey(str(__new_tz))
        )

    def as_utc(self) -> Timestamp:
        """convenience create a Timestamp instance in UTC timezone"""
        return self.with_tz(TimeZoneKey("UTC"))

    def to_dt_obj(
            self,
            tz_aware: bool = True
    ) -> dt.datetime:
        """
        create a python stdlib datetime object from this Timestamp,

        Args:
            tz_aware: if True, we'll create a tz aware datetime object, which means it has the tz info.
                default is True.

        Notes:
            Even if we specify tz_aware=False, the resulting python datetime object will be localized to the tz info
            on this Timestamp instance.
            See the warning here for why you probably want it this way:
                https://docs.python.org/3.9/library/datetime.html#datetime.datetime.utctimetuple
        """
        res: dt.datetime
        if tz_aware:
            utc_dt_obj = dt.datetime.fromtimestamp(
                self.to_python_timestamp(), tz=TimeZoneKey("UTC").as_py_zone_info()
            )

            if self.tz != TimeZoneKey("UTC"):
                res = utc_dt_obj.astimezone(self.tz.as_py_zone_info())
            else:
                res = utc_dt_obj
        else:
            res = dt.datetime.fromtimestamp(
                self.to_python_timestamp(), tz=None
            )
        return res

    def to_iso_str(
            self,
            tz_aware: bool = True
    ) -> str:
        """
        returns the iso formatted str, via the datetime method.
        see: https://docs.python.org/3.9/library/datetime.html#datetime.datetime.isoformat

        Args:
            tz_aware: if True, the string will contain the utc offset.

        Returns:
            str: formatted string
        """
        dt_obj = self.to_dt_obj(tz_aware=tz_aware)
        return dt_obj.isoformat()

    def to_str(
            self,
            fmat: str = "%x %X %Z",
    ) -> str:
        """
        create a string using the specified python format. see python docs for help with the codes:
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes

        Args:
            fmat: the python format, default is a human-readable format: "%x %X %Z"

        Returns: a formatted str
        """
        return self.to_dt_obj().strftime(fmat)

    def to_file_path_fmat(self) -> str:
        """
        convenience method to format the timestamp as a filepath friendly string, "%Y_%m_%dY%H_%M_%S_%f"
        """
        return self.to_str("%Y_%m_%dY%H_%M_%S_%f")

    def special_repr(self) -> str:
        return f"{self.__class__.__name__}['{self.to_str()}']"

    @classmethod
    def _from_datetime_datetime(
            cls,
            __obj: dt.datetime,
    ) -> Timestamp:
        """
        build the Timestamp instance from the python datetime object. If timezone specified, we'll use it otherwise
        we'll assume local as is the standard behavior
        """
        tz_key: TimeZoneKey
        if __obj.tzinfo:
            obj_tz_info: dt.tzinfo = __obj.tzinfo

            # zone info implementation vs other, pops up with
            #   the EST vs EDT (Eastern Standard Time, Eastern Daylight Savings Time).
            #  TODO: revisit this
            if isinstance(obj_tz_info, zoneinfo.ZoneInfo):
                tz_key = TimeZoneKey(obj_tz_info.key)
            else:
                tz_key = TimeZoneKey(obj_tz_info.tzname(__obj))
        else:
            tz_key = TimeZoneKey.local()
        return Timestamp(
            microseconds=int(__obj.timestamp() * 1e6),
            tz=tz_key
        )

    @classmethod
    def _from_datetime_date(
            cls,
            __obj: dt.date,
    ) -> Timestamp:
        datetime_obj = dt.datetime(year=__obj.year, month=__obj.month, day=__obj.day)
        return cls._from_datetime_datetime(datetime_obj)

    @classmethod
    def from_datetime(cls, __obj: t.Union[dt.datetime, dt.date]) -> Timestamp:
        """parse a stdlib representation of the date"""
        if isinstance(__obj, dt.datetime):
            return cls._from_datetime_datetime(__obj)
        elif isinstance(__obj, dt.date):
            return cls._from_datetime_date(__obj)
        else:
            raise ValueError(
                f"'{type(__obj)}' not one of the standard lib datetime types we support parsing"
            )

    @pydantic.model_validator(mode="before")
    @classmethod
    def _parse_common(cls, data: t.Any) -> t.Any:
        if isinstance(data, (dt.datetime, dt.date)):
            data = cls.from_datetime(data).model_dump()

        if isinstance(data, str):
            try:
                dt_obj = dt.datetime.fromisoformat(data)

            except Exception as e:
                raise ValueError(
                    f"failed to parse the string, '{data}', as iso formatted date str."
                ) from e

            data = cls.from_datetime(dt_obj).model_dump()
        return data

    @classmethod
    def now(
            cls,
            unit: t.Literal["us", "ms", "s"] = "ms",
            tz_key: t.Optional[t.Union[str, TimeZoneKey]] = None
    ) -> Timestamp:
        """
        get current date and time with local timezone, down to the specified unit

        Args:
            unit: the smallest unit to go down to.
                us: microsecond, 1 second = 1_000_000 microseconds
                ms: millisecond, 1 second = 1_000 milliseconds
                s: second, 1 second = 1 second
            tz_key: which timezone to use, if not specified, we'll use local

        Returns:
            a Timestamp instance
        """
        microseconds = int(1e6 * dt.datetime.utcnow().timestamp())

        data: int
        if unit == "us":
            data = microseconds
        elif unit == "ms":
            data = int((microseconds // 1e3) * 1e3)
        elif unit == "s":
            data = int((microseconds // 1e6) * 1e6)
        else:
            raise ValueError(f"'{unit}' is not a supported value")

        tz: TimeZoneKey
        if tz_key is not None:
            tz = TimeZoneKey(tz_key)
        else:
            tz = TimeZoneKey.local()

        return Timestamp(
            microseconds=data, tz=tz
        )

    @classmethod
    def utcnow(cls, unit: t.Literal["us", "ms", "s"] = "ms") -> Timestamp:
        """
        get current date and time with utc timezone, down to the specified unit

        Args:
            unit: the smallest unit to go down to.
                us: microsecond, 1 second = 1_000_000 microseconds
                ms: millisecond, 1 second = 1_000 milliseconds
                s: second, 1 second = 1 second

        Returns:
            a Timestamp instance
        """
        return cls.now(unit=unit, tz_key="UTC")

    @classmethod
    def today(cls) -> Timestamp:
        """get the current date, as a Timestamp object."""
        return Timestamp.model_validate(dt.datetime.today())

    def __get_logger_data__(self) -> _logging.log_extra.ExtraLogData:
        return _logging.log_extra.ExtraLogData(
            key="timestamp",
            desc="a timestamp object",
            data=self.special_repr()
        )

    def __eq__(self, other: t.Union[dt.datetime, dt.date, Timestamp]) -> bool:
        other_ts = Timestamp.model_validate(other)
        return self.microseconds == other_ts.microseconds

    def __gt__(self, other: t.Union[dt.datetime, dt.date, Timestamp]) -> bool:
        other_ts = Timestamp.model_validate(other)
        return self.microseconds > other_ts.microseconds



from spice_rack._bases import (
    _base_base as base_base,
    _mixins as mixins,
    _root as root,
    _value_model as value_model,
    _dispatchable as dispatchable,
    _special_str as special_str,
    _settings as settings,
    _exception as exceptions
)


# common ones here
PydanticBase = base_base.PydanticBase
ValueModelBase = value_model.ValueModelBase
DispatchableValueModelBase = dispatchable.DispatchableValueModelBase
RootModel = root.RootModel
SpecialStrBase = special_str.SpecialStrBase
SettingsBase = settings.SettingsBase
ClassId = dispatchable.ClassId



from __future__ import annotations
import json
import typing as t
from pathlib import Path
import pydantic
import pydantic_settings

from spice_rack._bases._settings import _sources


__all__ = ("SettingsBase", )


# PyCharm IDE works if we do this rather than Self from std lib
Self = t.TypeVar("Self", bound="SettingsBase")


class SettingsBase(pydantic_settings.BaseSettings):
    """
    Base class inheriting from v1 version of Pydantic's BaseSettings.
    This will be adjusted when we fully move to Pydantic v2.

    Notes:
        use double underscores, `__`, as the nested delimiter is default behavior

    References:
        - `pydantic-settings docs <https://docs.pydantic.dev/1.10/usage/settings/>`_
    """
    _singleton_source: t.ClassVar[t.Optional[_sources.SingletonSource]] = None

    model_config = pydantic_settings.SettingsConfigDict(
        extra="ignore"
    )

    def __init_subclass__(cls, singleton_mode: bool = True, **kwargs):
        if singleton_mode:
            cls._singleton_source = _sources.SingletonSource(
                settings_cls=cls
            )

        super().__init_subclass__(**kwargs)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[pydantic_settings.BaseSettings],
        init_settings: pydantic_settings.PydanticBaseSettingsSource,
        env_settings: pydantic_settings.DotEnvSettingsSource,
        dotenv_settings: pydantic_settings.PydanticBaseSettingsSource,
        file_secret_settings: pydantic_settings.PydanticBaseSettingsSource,
    ) -> tuple[pydantic_settings.PydanticBaseSettingsSource, ...]:

        sources: list[pydantic_settings.PydanticBaseSettingsSource] = [
            init_settings,
            env_settings,
            dotenv_settings,
            file_secret_settings
        ]
        if cls._singleton_source:
            sources.insert(1, cls._singleton_source)

        # replace with our extended func
        for src in sources:
            src.decode_complex_value = _decode_complex_value

        return tuple(sources)

    @classmethod
    def _get_secrets_dir(cls) -> t.Optional[t.Union[str, Path]]:
        """
        overwrite this to specify a directory to use for secrets, docker style.
        default is None.

        Notes:
            docker's target mount point is '/run/secrets/' in case you want that.
        """
        return None

    @classmethod
    def _get_dot_env_files(cls) -> t.List[t.Union[str, Path]]:
        """
        return relative or absolute file paths to .env files in descending precedence, so last
        one specified will be the highest precedence.
        """
        return []

    @classmethod
    def _get_env_nested_delimiter(cls) -> str:
        """
        the delimiter to use when parsing env values. default is double underscore
        """
        return "__"

    @classmethod
    def _get_env_prefix(cls) -> t.Optional[str]:
        """
        overwrite this to specify a prefix for all env variables this settings class will read.
        default is None.

        Notes:
            If you specify a value here, this will not include a delimiter like you may be expecting. A common
                recipe here would be to append the `_get_env_nested_delimiter' to the prefix you are specifying
        """
        return None

    @classmethod
    def load(
            cls: t.Type[Self],
            use_cls_env_files: bool = True,
            **kwargs
    ) -> Self:
        env_files: list[str] = []
        for fp in cls._get_dot_env_files():  # type: ignore
            env_files.append(str(fp))

        if use_cls_env_files:
            kwargs["_env_file"] = tuple(env_files)
            kwargs["_env_nested_delimiter"] = cls._get_env_nested_delimiter()
            kwargs["_env_prefix"] = cls._get_env_prefix()
        return cls(**kwargs)


# noinspection PyUnusedLocal
def _decode_complex_value(
        field_name: str,
        field: pydantic.fields.FieldInfo,
        value: t.Any
) -> t.Any:
    """
    overwrite pydantic's 'PydanticBaseSettingsSource' method to allow us to parse complex
    fields that have simple values. pydantic always tries to decode the data as json, even if it isn't.
    """
    try:
        return json.loads(value)

    # if it fails, we just return value directly
    except json.decoder.JSONDecodeError:
        return value

    except Exception as e:
        raise e



from __future__ import annotations
import typing as t
import polars as pl

import pydantic_core

__all__ = (
    "build_choices",
    "discrim_func"
)

_ChoicesT = t.Literal["df", "lazy_df", "row_dicts", "columnar_dict"]


def build_choices(
        df_schema: pydantic_core.CoreSchema,
        lazy_df_schema: pydantic_core.CoreSchema,
        row_dicts_schema: pydantic_core.CoreSchema,
        columnar_dict_schema: pydantic_core.CoreSchema
) -> t.Dict[_ChoicesT, pydantic_core.CoreSchema]:
    return {
        "df": df_schema,
        "lazy_df": lazy_df_schema,
        "row_dicts": row_dicts_schema,
        "columnar_dict": columnar_dict_schema,
    }


def discrim_func(_obj: t.Any) -> _ChoicesT:
    if isinstance(_obj, pl.DataFrame):
        return "df"
    elif isinstance(_obj, pl.LazyFrame):
        return "lazy_df"
    elif isinstance(_obj, list):
        return "row_dicts"
    elif isinstance(_obj, dict):
        return "columnar_dict"
    else:
        raise ValueError(f"unable to parse '{type(_obj)}' as a polars dataframe")



from __future__ import annotations
import typing as t

if t.TYPE_CHECKING:
    from spice_rack._fs_ops import _path_strs


__all__ = (
    "get_placeholder_rel_path",
    "is_placeholder_file_path",
)


def get_placeholder_rel_path() -> _path_strs.RelFilePathStr:
    from spice_rack._fs_ops import _path_strs
    return _path_strs.RelFilePathStr(
        "PLACEHOLDER_7b86519eb4ef4ef5b1e16aeb70329fd6.txt"
    )


def is_placeholder_file_path(path: _path_strs.RelOrAbsFilePathT) -> bool:
    placeholder_file_path = get_placeholder_rel_path()
    if str(path).endswith(str(placeholder_file_path)):
        return True
    else:
        return False



from __future__ import annotations


__all__ = (
    "get_current_datetime_for_file_path",
)


def get_current_datetime_for_file_path() -> str:
    from spice_rack._ts_service._timestamp import Timestamp
    return Timestamp.now().to_file_path_fmat()



from __future__ import annotations
from google.auth import credentials  # noqa

from spice_rack._gcp_auth._auth_strategies import _base


__all__ = (
    "AnonAuthStrategy",
)


class AnonAuthStrategy(_base.AbstractGcpAuthStrategy):
    """
    Authenticate to the client as anonymous.
    """

    def build_gcp_creds(self) -> credentials.Credentials:
        return credentials.AnonymousCredentials()



from spice_rack._ts_service._tz_key import *
from spice_rack._ts_service._timestamp import *
from spice_rack._ts_service._fields import *
from spice_rack._ts_service import _utils as utils, _validators as validators, _exts as exts





from __future__ import annotations
import typing as t
from abc import abstractmethod
from pydantic import Field
from spice_rack import _bases
from spice_rack._logging._log_level import LogLevel

if t.TYPE_CHECKING:
    from loguru import Logger


__all__ = (
    "AbstractLogSink",
)


class AbstractLogSink(
    _bases.dispatchable.DispatchableValueModelBase,
):
    level: LogLevel = Field(
        description="the level for this logger", default=LogLevel("info")
    )
    backtrace: bool = Field(
        description="if true, we automatically include tracebacks"
                    " when we are logging an error. We can still augment our "
                    "logs with a call stack whenever we want, "
                    "regardless of if we are in the exception context. "
                    "see the relevant log augmentor"
                    "see `docs <https://loguru.readthedocs.io/en/stable/overview.html"
                    "#fully-descriptive-exceptions>`_",
        default=True
    )
    diagnose: bool = Field(
        description="if we true, we annotate the stacktrace with values of specific variables"
                    " in the stack trace. Like as debugger. "
                    "see `docs <https://loguru.readthedocs.io/en/stable/overview.html"
                    "#fully-descriptive-exceptions>`_",
        default=False
    )
    struct_log: bool = Field(
        description="if true, we serialize the logs for usage in a structured way. "
                    "this field is used when building the loguru_kwargs "
                    "passed in when setting up the sink",
        default=False
    )
    include_extra_data: bool = Field(
        description="if false, we do not include extra data for this sink", default=True
    )
    custom_loguru_fmat: t.Optional[str] = Field(
        description="an optional field you can use to manually specify the loguru format for this "
                    "sink",
        default=None
    )

    def special_repr(self) -> str:
        """get a pretty str showing simple info about this logger"""
        return f"LoggerSink[type={self.class_id}]"

    def _get_default_loguru_format(self) -> str:
        """
        the loguru fmat to use for the logger. see: loguru docs on this.
        you can also overwrite this method to customize how a specific sink builds
        their default format str

        Returns:
            str: a loguru format
        """
        time_section = "<green>{time:YYYY-MM-DD at HH:mm:ss}</green>"
        level_section = "<blue>{level: <8}</>"
        service_name_section = "<magenta>{extra[service_name]: <15}</magenta>"
        line_num_section = "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>"
        message_section = "<green>{message}</>"
        extra_data_section = "<yellow>extra data: {extra[" + "extra_data" + "]}</yellow>"

        sections = [
            time_section, level_section, service_name_section, line_num_section,
            message_section
        ]

        if self.include_extra_data:
            sections.append(extra_data_section)

        format_str = " | ".join(sections)
        return format_str

    def get_loguru_format(self) -> str:
        if self.custom_loguru_fmat is not None:
            return self.custom_loguru_fmat
        else:
            return self._get_default_loguru_format()

    def _get_loguru_kwargs(self) -> dict:
        """build a dict of loguru setup kwargs from this sink"""
        return {
            "level": str(self.level).upper(),
            "format": self.get_loguru_format(),
            "serialize": self.struct_log,
            "backtrace": self.backtrace,
            "diagnose": self.diagnose
        }

    @abstractmethod
    def setup(self, logger: Logger, **custom_loguru_kwargs) -> None:
        raise NotImplementedError()



from __future__ import annotations
import typing as t
import yaml
import pydantic

from spice_rack._fs_ops import (
    _path_strs,
    _exceptions,
    _file_systems,
    _open_modes,
    _file_info,
)

from spice_rack._fs_ops._fs_models._base import AbstractFileSystemObj


__all__ = (
    "FilePath",
    "TextFilePath",
    "JsonFilePath",
    "YamlFilePath",
)


SelfTV = t.TypeVar("SelfTV", bound="_FilePathBase")


class _FilePathBase(AbstractFileSystemObj, class_type="root"):
    """
    implements standard file functionality. This is implemented as a 'root' to allow us to continue subclassing
    this for special file types.
    """

    path: _path_strs.AbsoluteFilePathStr = pydantic.Field(
        description="the path string on the underlying file system. Must be absolute. You can use the "
                    "file system-specific format, e.g. 'gs://bucket/file.txt' and we will convert it to "
                    "/bucket/file.txt' and make sure the file_system aligns"
    )

    def _special_repr_short(self) -> str:
        return f"File[path='{self.path}', file_system_type={self.file_system.special_repr()}]"

    def _special_repr_verbose(self) -> str:
        return f"File[path='{self.path}', file_system_type={self.file_system.special_repr()}]"

    def delete(
            self,
            if_non_existent: t.Literal["raise", "return"] = "return",
    ) -> None:
        try:
            return self.file_system.delete_file(
                self.path,
                if_non_existent=if_non_existent
            )
        except Exception as e:
            raise e

    def open(
            self,
            mode: _open_modes.SupportedOpenModesT = "rb"
    ) -> _open_modes.OpenFileT:
        try:
            return self.file_system.open_file(
                self.path,
                mode=mode
            )
        except Exception as e:
            raise e

    def write(self, data: t.Union[str, bytes], mode: t.Literal["wb", "ab"] = "wb") -> None:
        """
        convenience method to write bytes or str data to a file
        """
        byte_data: bytes
        if isinstance(data, bytes):
            byte_data = data
        else:
            byte_data = data.encode()

        with self.open(mode) as f:
            f.write(byte_data)

    def read_as_str(self, encoding: str = "utf-8") -> str:
        """convenience method to read str data from a file"""
        with self.open("rb") as f:
            byte_data = f.read()
        return byte_data.decode(encoding)

    def get_name(self, include_suffixes: bool = False) -> str:
        """get the name of the file, optionally stripping the name of the suffixes"""
        return self.path.get_name(include_suffixes=include_suffixes)

    def download_locally(
            self,
            dest_dir: _path_strs.AbsoluteDirPathStr
    ) -> _path_strs.AbsoluteFilePathStr:
        """download the file to the local dir specified"""
        return self.file_system.download_file_locally(
            self.path,
            dest_dir
        )

    def ensure_correct_file_ext(self, choices: list[str]) -> None:
        """
        ensure the file path has one of the specified extensions
        Args:
            choices: the valid extensions

        Returns: Nothing

        Raises:
            FilePathInvalidException: if the file path has no extension of it isn't
                one of the choices
        """
        return self.path.ensure_correct_file_ext([_file_info.FileExt(c) for c in choices])

    def ensure_correct_mime_type(self, choices: t.List[_file_info.MimeType]) -> None:
        """
        ensure the file path is one of the specified mime types
        Args:
            choices: the valid mime types

        Returns: Nothing

        Raises:
            InvalidFileMimeTypeException: if we cannot determine the mime type, or it is not one
                of the specified choices
        """
        return self.path.ensure_correct_mime_type([_file_info.MimeType(c) for c in choices])

    def get_file_ext(self) -> t.Optional[_file_info.FileExt]:
        """get the file extension from the file path if there is one"""
        return self.path.get_file_ext()

    def get_mime_type(self) -> t.Optional[_file_info.MimeType]:
        """get the mime type of the file from its file extension, if we have it"""
        return self.path.get_mime_type()

    @classmethod
    def init_from_str(cls: t.Type[SelfTV], raw_str: str) -> SelfTV:
        """
        parse a string into a FilePath instance, inferring the file system from the
        prefix of the string path.

        If the str starts with a '$' we will parse it as a DeferredFilePath instance and
        evaluate it right way into a FilePath inst.
        """
        if raw_str.startswith("$"):
            from spice_rack._fs_ops._fs_models._deferred import DeferredFilePath
            return DeferredFilePath.model_validate(raw_str).evaluate()
        else:
            inferred_fs = _file_systems.infer_file_system(raw_str)
            path_str = inferred_fs.clean_raw_path_str(raw_str)
            file_path = _path_strs.AbsoluteFilePathStr(path_str)
            return cls(path=file_path, file_system=inferred_fs)


class FilePath(_FilePathBase):
    """
    general file path, can have any file extension.
    """
    ...


class TextFilePath(_FilePathBase):
    """
    extension of the standard FilePath object that has json-specific methods.
    """
    @pydantic.model_validator(mode="before")
    def _handle_general_file_path(cls, data: t.Any) -> t.Any:
        if isinstance(data, FilePath):
            data = data.model_dump(exclude={"class_id"})
        return data

    def _post_init_validation(self) -> None:
        try:
            self.ensure_correct_file_ext(["txt"])

        except _exceptions.InvalidFileExtensionException as e:
            raise e.as_pydantic_error()

        except Exception as e:
            raise e

    def read_lines(self) -> t.List[str]:
        """read lines of the text file"""
        return self.read_as_str().splitlines()


_json_type_adapter = pydantic.TypeAdapter(pydantic.JsonValue)
"""use this to dump json data into data for writing"""


class JsonFilePath(_FilePathBase):
    """
    extension of the standard FilePath object that has json-specific methods.
    """
    @pydantic.model_validator(mode="before")
    def _handle_general_file_path(cls, data: t.Any) -> t.Any:
        if isinstance(data, FilePath):
            data = data.model_dump(exclude={"class_id"})
        return data

    def _post_init_validation(self) -> None:
        try:
            self.ensure_correct_file_ext(["json"])

        except _exceptions.InvalidFileExtensionException as e:
            raise e.as_pydantic_error()

        except Exception as e:
            raise e

    def json_write(self, obj: t.Any) -> None:
        """
        write the json-encodeable object. If a pydantic BaseModel, we'll use that to serialize

        Args:
            obj: any type of data we want to write, must be json-encodeable

        Returns:
            nothing

        Raises:
            PydanticValidationError: if the obj is not json-encodeable
        """
        byte_data: bytes
        if isinstance(obj, pydantic.BaseModel):
            byte_data = obj.model_dump_json().encode()

        else:
            # already bytes
            byte_data = _json_type_adapter.dump_json(
                _json_type_adapter.validate_python(obj)
            )
        self.write(byte_data, "wb")

    def json_read(self) -> pydantic.JsonValue:
        data = self.read_as_str()
        return _json_type_adapter.validate_json(data)


class YamlFilePath(_FilePathBase):
    """
    extension of the standard FilePath object that has yaml-specific methods.
    """
    @pydantic.model_validator(mode="before")
    def _handle_general_file_path(cls, data: t.Any) -> t.Any:
        if isinstance(data, FilePath):
            data = data.model_dump(exclude={"class_id"})
        return data

    def _post_init_validation(self) -> None:
        try:
            self.ensure_correct_file_ext(["yaml", "yml"])

        except _exceptions.InvalidFileExtensionException as e:
            raise e.as_pydantic_error()

        except Exception as e:
            raise e

    def yaml_write(self, obj: t.Any) -> None:
        encodeable_data: pydantic.JsonValue
        if isinstance(obj, pydantic.BaseModel):
            encodeable_data = obj.model_dump(mode="json")
        else:
            # already bytes
            encodeable_data = _json_type_adapter.dump_python(
                _json_type_adapter.validate_python(obj),
                mode="json"
            )
        with self.open("wb") as f:
            yaml.dump(encodeable_data, f, encoding="utf-8")

    def yaml_read(self) -> pydantic.JsonValue:
        with self.open("rb") as f:
            data = yaml.unsafe_load(f)

        return _json_type_adapter.validate_python(data)



from __future__ import annotations
import typing as t

from flatten_dict import flatten


__all__ = (
    "flatten_dict",
)


# restrict support for non-string based key reducers to allow us to enforce type annotation on the key


_REDUCER_FUNC_T = t.Callable[[t.Optional[t.Hashable], t.Hashable], str]
"""
type ann for a key reducer for the flattener. I inferred this from the source code here
https://github.com/ianlini/flatten-dict/blob/master/src/flatten_dict/reducers.py
"""


class _CustomSepReducerFunc:
    def __init__(self, sep: str):
        self.sep = sep

    def __call__(self, k1: t.Optional[t.Hashable], k2: t.Optional[t.Hashable]) -> str:
        pieces: t.List[str] = []
        if k1:
            pieces.append(str(k1))
            pieces.append(self.sep)
        pieces.append(str(k2))
        return "".join(pieces)


def flatten_dict(
        __dict: t.Dict[t.Any, t.Any],
        *,
        key_join_strat: t.Union[str, _REDUCER_FUNC_T] = "__",
        enumerate_types: t.Iterable[type] = (list, ),
        max_flatten_depth: t.Optional[int] = None,
        **kwargs
) -> t.Dict[str, t.Any]:
    """
    flatten the dict using https://github.com/ianlini/flatten-dict?tab=readme-ov-file#flatten

    Args:
        __dict: the dict to flatten
        key_join_strat: the strategy used to join keys when flattening, if a string is specified we treat is as
            shorthand for a func that joins keys using the str as the separator
        enumerate_types: an iterable of we want to enumerate, see doctest for example
        max_flatten_depth: specify this to restrict the depth we will flatten to. default is None, i.e. no restriction
        **kwargs: any other kwargs to pass through to https://github.com/ianlini/flatten-dict?tab=readme-ov-file#flatten

    Returns:
        a flattened dict

    >>> flatten_dict({"a": 1, "b": {"b": 1}})
    {'a': 1, 'b__b': 1}

    >>> flatten_dict({1: "a", 2: {2: "b"}})
    {'1': 'a', '2__2': 'b'}

    >>> flatten_dict({"a": ["x", "y", "z"]})
    {'a__0': 'x', 'a__1': 'y', 'a__2': 'z'}

    >>> flatten_dict({"a": ["x", "y", "z"]}, enumerate_types=())
    {'a': ['x', 'y', 'z']}
    """
    reducer_func: _REDUCER_FUNC_T
    if callable(key_join_strat):
        reducer_func = key_join_strat
    else:
        reducer_func = _CustomSepReducerFunc(sep=key_join_strat)
    return flatten(
        __dict,
        reducer=reducer_func,
        enumerate_types=enumerate_types,
        max_flatten_depth=max_flatten_depth,
        **kwargs
    )



from spice_rack._logging._log_level import *
from spice_rack._logging._logger import *
from spice_rack._logging._logger_getter import *
from spice_rack._logging import (
    _sinks as sinks,
    _log_aug as log_extra
)
LoggableObjMixin = log_extra.LoggableObjMixin
